"""Testing nodes with migen, a.k.a. nodes with PyMigenBody."""

import io
import json
import logging
import unittest
import unittest.mock

from test._utils import InputCheckerWithExit, add_non_const

import dill
import migen

from deltalanguage.data_types import BaseDeltaType, DOptional
from deltalanguage.lib import StateSaver
from deltalanguage.runtime import (DeltaPySimulator,
                                   deserialize_graph,
                                   serialize_graph)
from deltalanguage.wiring import (DeltaBlock,
                                  DeltaGraph,
                                  MigenNodeTemplate,
                                  PythonBody)

from examples.tutorials.migen_hardware_examples import (
    generate_graph_constant_input,
    generate_graph_interactive_input
)

from test._utils import assert_capnp_content_types


class AlternatingOutputsMigen(MigenNodeTemplate):
    """
    This node sends a counter value to both output ports
    but only asserts the valid for one at the time
    """

    def migen_body(self, template):
        start = template.add_pa_in_port('start', DOptional(int))
        out_a = template.add_pa_out_port('out_a', int)
        out_b = template.add_pa_out_port('out_b', int)
        # This will need to be converted to boolean when migen nodes support
        # boolean
        self.cnt = migen.Signal(10)

        self.comb += (
            out_a.ready.eq(1),
            out_b.ready.eq(1),
            start.ready.eq(1)
        )
        self.sync += migen.If(
            self.cnt & 0x1,
            out_a.valid.eq(start.data),
            out_b.valid.eq(0)
        ).Else(
            out_a.valid.eq(0),
            out_b.valid.eq(start.data)
        )
        self.sync += (
            self.cnt.eq(self.cnt+1),
            out_a.data.eq(self.cnt),
            out_b.data.eq(self.cnt)
        )


@DeltaBlock(allow_const=False)
def adder(a: int, b: int) -> int:
    return a + b


@DeltaBlock(allow_const=False)
def multiplier(a: int) -> int:
    return a*1000


class TestMigen(MigenNodeTemplate):
    """Example class should subclass MigenNodeTemplate and define migen_body.
    """

    def migen_body(self, template):
        # Input/Outputs start here:

        # 2 inputs and 2 outputs.
        #
        # This block group ports which will be accessed by migen, using
        # the protocol adapters.
        # in_ports and out_ports implement a similar logic based on 3
        # signals, ready, valid and data.
        # An input can be received when the ready signal is = '1'.
        # data contains the value of the message that we are receiving
        # and can considered sensible only when valid = '1', i.e. when
        # a new data has been received on the pa_input_port.
        # The opposite logic holds true for the outputs.
        in1 = template.add_pa_in_port('in1', DOptional(int))
        in2 = template.add_pa_in_port('in2', DOptional(int))
        out1 = template.add_pa_out_port('out1', int)
        out2 = template.add_pa_out_port('out2', int)

        # The main migen logic starts here:

        # Everything below is just an example that show different routines.

        # Add a 32-bit counter (0-2**32-1) which will increment at each clock
        # cycle.
        self.counter = migen.Signal(32)
        self.sync += self.counter.eq(self.counter + 1)

        # Add a condition when in_ports are ready.
        self.comb += migen.If(
            self.counter >= 3,
            in1.ready.eq(1),
            in2.ready.eq(1)
        )

        # Pretend that we do a useful calculations.
        # Here we first check that the outputs are ready.
        # Then wait for the counter to reach 100.
        # And write outputs.
        # Note that the output should be marked as valid.
        self.comb += migen.If(
            (out1.ready & out2.ready) == 1,
            migen.If(
                self.counter == 5,
                out1.data.eq(in1.data + in2.data),
                out2.data.eq(self.counter),
                out1.valid.eq(in1.valid & in2.valid),
                out2.valid.eq(in1.valid & in2.valid)
            ).Else(
                out1.valid.eq(0),
                out2.valid.eq(0)
            )
        )


class MigenNodeSerialisationTest(unittest.TestCase):

    def setUp(self):
        DeltaGraph.clean_stack()

    def test_serialisation(self):
        """Serialize/deserialize a graph with a node with a PyMigenBody.

        Notes
        -----
        The content of the bodies depends on the environment, i.e. how the test
        is executed. For this reason we just compare the structure of the graph
        here.
        """
        s = StateSaver(int)

        example_migen = TestMigen(tb_num_iter=2000,
                                  name='counter',
                                  lvl=logging.INFO,
                                  vcd_name="/workdir/TestMigen.vcd")

        with DeltaGraph() as graph:
            example_migen_out = example_migen.call(in1=40, in2=2)
            s.save_and_exit(adder(example_migen_out.out1,
                                  multiplier(example_migen_out.out2)))

        data, _ = serialize_graph(graph)
        self.assertEqual(type(data), bytes)
        g_capnp = deserialize_graph(data).to_dict()
        assert_capnp_content_types(self, g_capnp)
        with open('test/data/graph_with_migen_capnp.json', 'r') as file:
            self.assertEqual(g_capnp, json.load(file))

    def test_one_migen_node_with_2_outs(self):
        """One PyMigenBody with 2 out ports produces what we expect."""
        s = StateSaver(int, verbose=True)

        with DeltaGraph() as graph:
            counter = TestMigen(tb_num_iter=2000,
                                name='counter',
                                lvl=logging.INFO).call(in1=40, in2=2)
            s.save_and_exit(adder(counter.out1, multiplier(counter.out2)))

        rt = DeltaPySimulator(graph)
        rt.run()

        self.assertEqual(s.saved, [5042])

    def test_one_migen_node_with_separate_ctrl_on_output_valid(self):
        """One PyMigenBody with 2 optional output ports produces
        the correct valid values (for the different ports).
        The migen node should be generating a sequence of outputs
        (1, None), (None, 2), (3, None) etc... """
        with DeltaGraph() as graph:
            alt = AlternatingOutputsMigen(tb_num_iter=100,
                                          name='alternatingOutput',
                                          lvl=logging.INFO).call(start=1)
            my_adder = add_non_const(alt.out_a, alt.out_b)
            # Checking that we have received a 1 and a 2
            checker = InputCheckerWithExit(lambda x: (x == 3))
            checker.check(my_adder)

        rt = DeltaPySimulator(graph)
        rt.run()

        self.assertEqual(checker.cond_met, True)


class MigenNodeHardwareBlocksTest(unittest.TestCase):

    def setUp(self):
        DeltaGraph.clean_stack()

    @unittest.mock.patch('sys.stdout', new_callable=io.StringIO)
    def test_migen_node_reset_shaper_constant(self, mock_stdout):
        """One PyMigenBody that takes as input a reset signal and extends the
        length of the reset to N clock cycles reset.
        """
        graph, _ = generate_graph_constant_input()
        rt = DeltaPySimulator(graph)
        rt.run()
        self.assertEqual(
            mock_stdout.getvalue(),
            'CHECK_SHAPE: reset has lasted at least 5 clk cycles\n'
        )

    @unittest.mock.patch('sys.stdout', new_callable=io.StringIO)
    def test_migen_node_reset_shaper_pulse(self, mock_stdout):
        """One PyMigenBody that takes as input a reset signal and extends the
        length of the reset to N clock cycles reset. We generate a pulse of
        length 1 clock cycle (i.e. 0010) and check for (00000...111110)
        """

        graph, _ = generate_graph_interactive_input(verbose=False)
        rt = DeltaPySimulator(graph)
        rt.run()
        self.assertEqual(mock_stdout.getvalue(),
                         'CHECK_SHAPE: reset has lasted exactly 5 clk cycles\n')


if __name__ == "__main__":
    unittest.main()



"""Test of DeltaGraph class."""

from typing import OrderedDict
import unittest

import deltalanguage as dl

from test._utils import return_1, TwoIntsT, TwoInts


@dl.DeltaBlock()
def union_one_receiver(x: dl.DUnion([int])) -> int:
    return x


ForkedReturnT, ForkedReturn = dl.make_forked_return({
    'a': int, 'b': bool, 'c': int, 'd': int
})


@dl.DeltaBlock()
def forked_return_output(x: dl.DInt(dl.DSize(8)), y: dl.DInt(dl.DSize(8))) -> ForkedReturnT:
    return ForkedReturn(a=0, b=1, c=1, d=0)


@dl.DeltaBlock()
def forked_return_output_no_input() -> ForkedReturnT:
    return ForkedReturn(a=0, b=1, c=1, d=0)


@dl.DeltaBlock(allow_const=False)
def multi_body_no_output(i: dl.DInt(dl.DSize(8))) -> dl.Void:
    print(i)


class Foo():

    def __init__(self):
        pass

    @dl.DeltaMethodBlock()
    def method_func_no_output(self, i: dl.DInt(dl.DSize(8))) -> dl.Void:
        print(i + 1)


class MigenFoo(dl.MigenNodeTemplate):
    def migen_body(self, template):
        template.add_pa_in_port('i', dl.DOptional(dl.DInt(dl.DSize(8))))


@dl.Interactive({'i': dl.DInt(dl.DSize(8))}, out_type=dl.Void)
def interactive_func_no_output(node: dl.RealNode):
    a = node.receive('i')


template_no_output_no_body = dl.NodeTemplate(
    name="template_no_output_no_body",
    in_params=OrderedDict([('i', dl.DInt(dl.DSize(8)))]),
    out_type=dl.Void
)


@dl.DeltaBlock()
def experiment_stopper(completed: dl.DInt(dl.DSize(8))) -> dl.Void:
    raise dl.DeltaRuntimeExit


class DeltaGraphTest(unittest.TestCase):
    """Check rules used by ``DeltaGraph.check``."""

    def test_loose_output(self):
        """Loose output wire don't cause a problem"""
        s = dl.lib.StateSaver(int)
        with dl.DeltaGraph() as graph:
            s.transfer(4)

        node = graph.find_node_by_name('transfer')
        self.assertTrue(len(node.out_ports) == 0)
        self.assertTrue(graph.check())

    def test_loose_input(self):
        """Loose input wire is bad."""
        s = dl.lib.StateSaver(int)
        with dl.DeltaGraph() as graph:
            p = dl.placeholder_node_factory()
            s.transfer(p)

        with self.assertRaises(dl.data_types.DeltaIOError):
            graph.check()

    def test_loose_optional_input(self):
        """Optional input without input does not raise an error."""
        @dl.DeltaBlock(allow_const=False)
        def test_node(a: dl.DOptional(int)) -> dl.Void:
            pass

        with dl.DeltaGraph() as graph:
            p = dl.placeholder_node_factory()
            test_node(p)

        self.assertTrue(graph.check())


def to_union_of_one(node: dl.RealNode) -> dl.RealNode:
    org_t = node.out_type
    node.out_type = dl.DUnion([org_t])

    return node


class DeltaGraphStrTest(unittest.TestCase):
    """Test for correct graph printing."""

    def test_graph_str_output(self):
        """Stringify a pre-defined graph and compare output to test data"""

        dl.DeltaGraph.clean_stack()

        with dl.DeltaGraph() as graph:
            n1 = forked_return_output(1, 2)
            n2 = template_no_output_no_body.call(n1.a)
            n3 = multi_body_no_output(n1.b)
            n4 = MigenFoo().call(i=n1.d)
            n5 = experiment_stopper(n1.c)

        obj = Foo()
        n3.add_body(obj.method_func_no_output)
        n3.add_body(interactive_func_no_output)

        with open('test/data/graph_str.txt', 'r') as file:
            correct_str = file.read()

        self.assertEqual(str(graph), correct_str)


class DeltaGraphWiringWrongInputTest(unittest.TestCase):
    """Error is raised when trying to provide an input to an in-port that
    is not defined.
    """

    def test_DeltaBlock(self):
        @dl.DeltaBlock()
        def foo(a: int) -> dl.Void:
            print(a)

        with self.assertRaises(dl.data_types.DeltaIOError):
            with dl.DeltaGraph():
                foo(b=1)

    def test_DeltaMethodBlock(self):
        class AClass:

            @dl.DeltaMethodBlock()
            def add_x(self, a: int) -> dl.Void:
                print(a)

        with self.assertRaises(dl.data_types.DeltaIOError):
            with dl.DeltaGraph():
                AClass().add_x(b=1)

    def test_Interactive(self):
        @dl.Interactive({'a': int}, dl.Void)
        def foo(node):
            node.receive("a")

        with self.assertRaises(dl.data_types.DeltaIOError):
            with dl.DeltaGraph():
                foo.call(b=1)

    def test_MigenNodeTemplate(self):
        class AMigenNode(dl.MigenNodeTemplate):
            def migen_body(self, template):
                template.add_pa_in_port('a', dl.DOptional(int))

        with self.assertRaises(dl.data_types.DeltaIOError):
            with dl.DeltaGraph():
                AMigenNode().call(b=1)


class DeltaGraphWiringNoOutputTest(unittest.TestCase):
    """Error is raised when trying to get an output from a node without defined
    out-ports.
    """

    def test_DeltaBlock(self):
        @dl.DeltaBlock()
        def foo(a: int) -> dl.Void:
            print(a)

        with self.assertRaises(dl.data_types.DeltaIOError):
            with dl.DeltaGraph():
                foo(foo(1))

    def test_DeltaMethodBlock(self):
        class AClass:

            @dl.DeltaMethodBlock()
            def foo(self, a: int) -> dl.Void:
                print(a)

        with self.assertRaises(dl.data_types.DeltaIOError):
            with dl.DeltaGraph():
                AClass().foo(AClass().foo(1))

    def test_Interactive(self):
        @dl.Interactive({"a": int}, dl.Void)
        def foo(node):
            print(node.receive("a"))

        with self.assertRaises(dl.data_types.DeltaIOError):
            with dl.DeltaGraph():
                foo.call(a=foo.call(a=1))

    def test_MigenNodeTemplate(self):
        class AMigenNode(dl.MigenNodeTemplate):
            def migen_body(self, template):
                template.add_pa_in_port('a', dl.DOptional(int))

        with self.assertRaises(dl.data_types.DeltaIOError):
            with dl.DeltaGraph():
                AMigenNode().call(a=AMigenNode().call(a=1))


class DeltaGraphWiringWrongOutputTest(unittest.TestCase):
    """Error is raised when trying to get an output from a node with multiple
    output ports, but use a wrong port name.
    """

    def test_DeltaBlock(self):
        @dl.DeltaBlock()
        def foo(a: int) -> TwoIntsT:
            return TwoInts(x=1, y=2)

        with self.assertRaises(dl.data_types.DeltaIOError):
            with dl.DeltaGraph():
                foo(foo(1).z)

    def test_DeltaMethodBlock(self):
        class AClass:

            @dl.DeltaMethodBlock()
            def foo(self, a: int) -> TwoIntsT:
                return TwoInts(x=1, y=2)

        with self.assertRaises(dl.data_types.DeltaIOError):
            with dl.DeltaGraph():
                AClass().foo(AClass().foo(1).z)

    def test_Interactive(self):
        @dl.Interactive({"a": int}, TwoIntsT)
        def foo(node):
            node.send(TwoInts(x=1, y=2))

        with self.assertRaises(dl.data_types.DeltaIOError):
            with dl.DeltaGraph():
                foo.call(a=foo.call(a=1).z)

    def test_MigenNodeTemplate(self):
        class AMigenNode(dl.MigenNodeTemplate):
            def migen_body(self, template):
                template.add_pa_in_port('a', dl.DOptional(int))
                template.add_pa_out_port('x', int)
                template.add_pa_out_port('y', int)

        with self.assertRaises(dl.data_types.DeltaIOError):
            with dl.DeltaGraph():
                AMigenNode().call(a=AMigenNode().call(a=1).z)


if __name__ == "__main__":
    unittest.main()



"""This submodule is dedicated to the runtime simulation as well as
universal routines used by Deltasimulator and Deltaruntime.

This submodule contains the following user-facing parts:

- Serialization routines:
  :py:func:`serialize_graph`
  :py:func:`deserialize_graph`
- Python Runtime Simulator :py:class:`DeltaPySimulator`

- Universal exit strategy :py:exc:`DeltaRuntimeExit`

For convenience the classes and functions listed above are also exposed to
the top level module and can be used like this:

.. code-block:: python

    import deltalanguage as dl

    print(dl.DeltaPySimulator())
    print(dl.DeltaRuntimeExit)
"""

from ._output import deserialize_graph, serialize_graph
from ._queues import ConstQueue, DeltaQueue
from ._runtime import DeltaPySimulator, DeltaRuntimeExit, DeltaThread


# user-facing classes
__all__ = ["deserialize_graph",
           "serialize_graph",
           "DeltaPySimulator",
           "DeltaRuntimeExit"]



"""Types that cannot actually be used as python types.
They do not integrate with mypy.
They do not have values you can interact with.
"""

from abc import ABC, abstractmethod
import logging
import struct
from typing import (Any,
                    Dict,
                    Iterable,
                    List,
                    NamedTuple,
                    Tuple,
                    Type,
                    Union)

import attr
import numpy as np

from deltalanguage.logging import make_logger

from ._exceptions import DeltaTypeError
from ._special import DSize, Void

logger = make_logger(logging.WARNING, "Data Types")


class BaseDeltaType(ABC):
    """The base Deltaflow type, of which other types are derived.

    Attributes
    ----------
    size : DSize
        Full size of the data type. There are several cases:

        * for primitive types it's the number of bits
        * for compound types it's the sum of all componenets' sizes
        * for :py:class:`DUnion` that is the the size of the largest element
          PLUS 1 byte that stores which particular data type is encoded
    """

    def __init__(self):
        self._pack_format = None
        self.size = DSize(0)

    def bytes_to_bits(self, s: bytes) -> bytes:
        """Converts a byte encoding of an object to a bit encoding.

        Parameters
        ----------
        s : bytes
            The object encoded as a byte string.

        Returns
        -------
        bytes
            The same object encoded as a bit string.
        """
        if s == "":
            return ""
        return format(int.from_bytes(s, byteorder="little"),
                      f'0{self.size}b').encode("ascii")

    def bits_to_bytes(self, s: bytes) -> bytes:
        """Converts a bit string encoding of an object to a byte string.

        Parameters
        ----------
        s : bytes
            Bit string encoding of an object.

        Returns
        -------
        bytes
            The same object encoded as a byte string.
        """
        return int(s.decode("ascii"), 2).to_bytes((len(s) + 7) // 8, 'little')

    @abstractmethod
    def __str__(self):
        pass

    @abstractmethod
    def as_python_type(self) -> Type:
        """Returns the equivalent Python data type.

        For compound data structures this method will be called
        recursively for all subcomponents.

        .. note::
            Python types do not support the idea of fixed size, thus this
            mapping is surjective, for instance ``DInt(DSize(32))`` and
            ``DInt(DSize(64))`` both are mapped to ``int``.
        """
        pass

    @abstractmethod
    def as_numpy_type(self) -> Type:
        """Returns an equivalent NumPy data type.

        For compound data structures this method will be called
        recursively for all subcomponents.

        .. note::
            NumPy types only support limited fixed sizes. For instance,
            ``DUInt(DSize(8))`` is compatible with ``numpy.uint8``,
            but ``DUInt(DSize(128))`` has no compatible numpy type.
        """
        pass

    @abstractmethod
    def as_numpy_object(self, val):
        """Returns a NumPy object set to val.

        For primitive delta types, this is the equivalent Python type.
        For compound types, this is a Numpy array whose elements are
        the corresponding types.
        """
        pass

    @abstractmethod
    def from_numpy_object(self, val):
        """Given a NumPy object, converts to an equivalent Python object.

        .. warning::
            Only one-dimensional NumPy arrays are currently supported.
        """
        pass

    def __repr__(self):
        return str(self)

    @abstractmethod
    def pack(self, val: object) -> bytes:
        """Encode an object to a bit string.

        For compound data structures this method will be called
        recursively for all subcomponents.

        Parameters
        ----------
        val : object
            Object to be packed.
            Its type should match :py:meth:`as_python_type`.

        Returns
        -------
        bytes
        """
        pass

    @abstractmethod
    def unpack(self, buffer: bytes) -> object:
        """Decode bit string to an object.

        Parameters
        ----------
        bytes
            Bytestring with a packed object of the class.

        Returns
        -------
        object
            Object itself.
        """
        pass

    @abstractmethod
    def is_packable(self, val: object) -> bool:
        """Checks if the object can be packed as this type.

        Parameters
        ----------
        val : object
            Object to be packed.

        Returns
        -------
        bool
            True if the object can be packed as this type, otherwise False.
        """

    @property
    def pack_format(self):
        """Format for :py:func:`struct.pack` defined by
        :py:meth:`set_pack_format`.

        Return
        ------
        str
        """
        if self._pack_format is None:
            self.set_pack_format()
        return self._pack_format

    @abstractmethod
    def set_pack_format(self):
        """Set :py:attr:`pack_format`.

        .. todo::
            This can be implemented via `@property.setter`.
        """
        pass


class PrimitiveDeltaType(BaseDeltaType):
    """Primitive Deltaflow types, which have a direct mapping to primitive
    types of Python and C.
    """

    def __eq__(self, other):
        if type(self) is type(other):
            return self.size == other.size
        return False

    def __hash__(self):
        return hash(self.size)

    def pack(self, val):
        return self.bytes_to_bits(struct.pack(self.pack_format, val))

    def unpack(self, buffer):
        return struct.unpack(self.pack_format, self.bits_to_bytes(buffer))[0]

    def as_numpy_object(self, val):
        return self.as_numpy_type()(val)

    def from_numpy_object(self, val):
        return self.as_python_type()(val)

    def is_packable(self, val):
        return delta_type(val) == self


class CompoundDeltaType(BaseDeltaType):
    """Compound Deltaflow types are constucted using primitive data types."""

    @abstractmethod
    def __len__(self):
        pass

    @abstractmethod
    def __hash__(self):
        pass

    @abstractmethod
    def flatten(self, vals) -> list:
        """Converts data of this type to a plain list."""
        pass

    @abstractmethod
    def deflatten(self, flatten_vals: list):
        """Reverts :py:meth:`flatten`. The return type is exactly the same as
        :py:meth:`as_python_type` for each data type.
        """
        pass


class Top(BaseDeltaType):
    """Default Deltaflow type. Used if type is not recognised.
    Defined as a singleton.
    """

    _instance = None

    # allocation for a generic object size
    size = DSize(100000)

    def __new__(cls):
        """If a Top object type already exists, return that type.
        Otherwise, generate a Top object type.
        Based on https://python-patterns.guide/gang-of-four/singleton/

        Returns
        -------
            A Deltaflow Top type.
        """
        if cls._instance is None:
            # Use BaseDeltaType __new__ function to construct a Top object.
            cls._instance = super(Top, cls).__new__(cls)
        return cls._instance

    def __str__(self):
        return "T"

    def as_python_type(self):
        return Any

    def as_numpy_type(self):
        raise NotImplementedError

    def as_numpy_object(self, val):
        raise NotImplementedError

    def from_numpy_object(self, val):
        raise NotImplementedError

    def pack(self, val):
        raise NotImplementedError

    def unpack(self, buffer):
        raise NotImplementedError

    def is_packable(self, val):
        raise NotImplementedError

    def set_pack_format(self):
        raise NotImplementedError


class DArray(CompoundDeltaType):
    """Fixed-size collection of objects of the same type.

    Parameters
    ----------
    list_of : Union[Type, BaseDeltaType]
        Type of elements. Note that elements can be of both primitive and
        compound types.
    length : DSize
        Number of elements.
    """

    def __init__(self, list_of: Union[Type, BaseDeltaType], length: DSize):
        if not isinstance(length, DSize):
            raise DeltaTypeError
        super().__init__()
        self.list_of = as_delta_type(list_of)
        self.length = length
        self.size = self.list_of.size * self.length

    def __str__(self):
        return f"[{self.list_of} x {self.length}]"

    def __len__(self):
        return self.length

    def as_python_type(self):
        # python does not have fixed size List
        return List[self.list_of.as_python_type()]

    def as_numpy_object(self, val):
        return np.array([self.list_of.as_numpy_object(v) for v in val],
                        dtype=self.list_of.as_numpy_type())

    def from_numpy_object(self, val):
        if type(val) is not np.ndarray:
            # Array is wrapped in a tuple object
            if len(val) != 1:
                logger.warning(
                    "Warning, NumPy array object for DArray has length != 1")
            val = val[0]
        return [self.list_of.from_numpy_object(v) for v in val]

    def as_numpy_type(self):
        return np.dtype([('f0',
                          self.list_of.as_numpy_type(),
                          self.length.val)])

    def __eq__(self, other):
        if type(self) is type(other):
            return self.list_of == other.list_of and self.length == other.length
        return False

    def __hash__(self):
        return hash((self.list_of, self.length))

    def flatten(self, vals):
        flatten_vals = []
        for v in vals:
            v_df_type = delta_type(v)
            if is_primitive(v_df_type):
                flatten_vals.append(v)
            else:
                flatten_vals += v_df_type.flatten(v)
        return flatten_vals

    def deflatten(self, flatten_vals):
        vals = []
        if is_primitive(self.list_of):
            for _ in range(self.length.val):
                # TODO pop from the end would be better
                vals.append(flatten_vals.pop(0))
        else:
            for _ in range(self.length.val):
                vals.append(self.list_of.deflatten(flatten_vals))

        return vals

    def pack(self, val):
        if not self.is_packable(val):
            raise DeltaTypeError(f'Data does not match the packing format\n'
                                 f'{val=}\n{self=}')

        if type(val) is np.ndarray:
            val = self.from_numpy_object(val)

        return b"".join([self.list_of.pack(v) for v in val])

    def unpack(self, buffer):
        return [self.list_of.unpack(buffer[i:i+self.list_of.size.val])
                for i in range(0, len(buffer), self.list_of.size.val)]

    def is_packable(self, val):
        if not isinstance(val, (List, np.ndarray)):
            return False
        elif len(val) == self.length.val:
            return all(self.list_of.is_packable(v) for v in val)
        else:
            return False

    def set_pack_format(self):
        self._pack_format = self.list_of.pack_format * self.length.val


class DUInt(PrimitiveDeltaType):
    """Unsigned integers with a given number of bits.

    Parameters
    ----------
    size : DSize
        Full size of the data type.
    """

    def __init__(self, size: DSize = DSize(32)):
        if not isinstance(size, DSize):
            raise DeltaTypeError
        super().__init__()
        self.size = size

    def __str__(self):
        return f"DUInt{self.size}"

    def as_python_type(self):
        return int

    def as_numpy_type(self):
        if self.size == DSize(8):
            return np.uint8
        elif self.size == DSize(16):
            return np.uint16
        elif self.size == DSize(32):
            return np.uint32
        elif self.size == DSize(64):
            return np.uint64
        else:
            raise NotImplementedError(f'Unsupported format: {self.size} - '
                                      'supported sizes are 8, 16, 32, 64')

    def set_pack_format(self):
        if self.size == DSize(8):
            self._pack_format = 'B'
        elif self.size == DSize(16):
            self._pack_format = 'H'
        elif self.size == DSize(32):
            self._pack_format = 'I'
        elif self.size == DSize(64):
            self._pack_format = 'Q'
        else:
            raise NotImplementedError(f'Unsupported format: {self.size} - '
                                      'supported sizes are 8, 16, 32, 64')


class DInt(PrimitiveDeltaType):
    """Signed integers with a given number of bits.

    Parameters
    ----------
    size : DSize
        Full size of the data type.
    """

    def __init__(self, size: DSize = DSize(32)):
        if not isinstance(size, DSize):
            raise DeltaTypeError
        super().__init__()
        self.size = size

    def __str__(self):
        return f"DInt{self.size}"

    def as_python_type(self):
        return int

    def as_numpy_type(self):
        if self.size == DSize(8):
            return np.int8
        elif self.size == DSize(16):
            return np.int16
        elif self.size == DSize(32):
            return np.int32
        elif self.size == DSize(64):
            return np.int64
        else:
            raise NotImplementedError(f'Unsupported format: {self.size} - '
                                      'supported sizes are 8, 16, 32, 64')

    def set_pack_format(self):
        if self.size == DSize(8):
            self._pack_format = 'b'
        elif self.size == DSize(16):
            self._pack_format = 'h'
        elif self.size == DSize(32):
            self._pack_format = 'i'
        elif self.size == DSize(64):
            self._pack_format = 'q'
        else:
            raise NotImplementedError(f'Unsupported format: {self.size} - '
                                      'supported sizes are 8, 16, 32, 64')


class DBool(PrimitiveDeltaType):
    """Boolean 1-bit type."""

    def __init__(self):
        super().__init__()
        self.size = DSize(1)

    def __str__(self):
        return "DBool"

    def as_python_type(self):
        return bool

    def as_numpy_type(self):
        return np.bool_

    def set_pack_format(self):
        self._pack_format = '?'

    def pack(self, val):
        if val == True:
            return b'1'
        elif val == False:
            return b'0'
        else:
            raise DeltaTypeError

    def unpack(self, buffer):
        if buffer == b'0':
            return False
        elif buffer == b'1':
            return True
        else:
            raise DeltaTypeError


class DChar(PrimitiveDeltaType):
    """Character implemented via 8-bit in analogy with ``char`` from ``struct``
    library.
    """

    def __init__(self):
        super().__init__()
        self.size = DSize(8)

    def __str__(self):
        return "DChar8"

    def as_python_type(self):
        raise NotImplementedError('Python does not have char type')

    def as_numpy_type(self):
        return np.uint8

    def as_numpy_object(self, val):
        return np.ubyte(ord(val))

    def from_numpy_object(self, val):
        return chr(val)

    def set_pack_format(self):
        self._pack_format = 'c'


class DStr(DArray):
    """Array of characters."""

    def __init__(self, length: DSize = DSize(1024)):
        if not isinstance(length, DSize):
            raise DeltaTypeError
        super().__init__(DChar(), length)

    def __str__(self):
        return f"DStr{self.size}"

    def as_python_type(self):
        return str

    def as_numpy_type(self):
        return (np.string_, self.length.val)

    def as_numpy_object(self, val):
        return np.string_(val)

    def from_numpy_object(self, val):
        return val.decode("ascii")

    def flatten(self, vals):
        return [vals.encode('ascii')]

    def deflatten(self, flatten_vals):
        val = flatten_vals.pop(0)
        return val.decode('ascii').rstrip('\x00')

    def pack(self, val):
        # overwrite
        if not self.is_packable(val):
            raise DeltaTypeError(f'Data does not match the packing format\n'
                                 f'{val=}\n{self=}')

        if type(val) is np.string_:
            val = self.from_numpy_object(val)
        return self.bytes_to_bits(struct.pack(self.pack_format,
                                              *self.flatten(val)))

    def unpack(self, buffer):
        # overwrite
        flatten_vals = list(struct.unpack(self.pack_format,
                                          self.bits_to_bytes(buffer)))
        return self.deflatten(flatten_vals)

    def is_packable(self, val):
        val_df_type = delta_type(val)
        return type(val_df_type) is DStr and val_df_type.length <= self.length

    def set_pack_format(self):
        self._pack_format = str(self.length) + 's'


class DFloat(PrimitiveDeltaType):
    """Floating point numbers.

    For the concrete implementation of encoding see ``struct``, in particular
    the packing uses IEEE 754 with a single integer number for the size.
    """

    def __init__(self, size: DSize = DSize(32)):
        if not isinstance(size, DSize):
            raise DeltaTypeError
        super().__init__()
        self.size = size

    def __str__(self):
        return f"DFloat{self.size}"

    def as_python_type(self):
        return float

    def as_numpy_type(self):
        if self.size == DSize(32):
            return np.float32
        elif self.size == DSize(64):
            return np.float64
        else:
            raise NotImplementedError(f'Unsupported format: {self.size} - '
                                      'supported sizes are 32, 64')

    def set_pack_format(self):
        if self.size == DSize(32):
            self._pack_format = 'f'
        elif self.size == DSize(64):
            self._pack_format = 'd'
        else:
            raise NotImplementedError(f'Unsupported format: {self.size} - '
                                      'supported sizes are 32, 64')


class DComplex(PrimitiveDeltaType):
    """Complex numbers are implemented as two floating point numbers
    for real and imaginary components.
    """

    def __init__(self, size: DSize = DSize(64)):
        if not isinstance(size, DSize):
            raise DeltaTypeError
        super().__init__()
        self.size = size

    def __str__(self):
        return f"DComplex{self.size}"

    def as_python_type(self):
        return complex

    def as_numpy_type(self):
        if self.size == DSize(64):
            return np.complex64
        elif self.size == DSize(128):
            return np.complex128
        else:
            raise NotImplementedError(f'Unsupported format: {self.size} - '
                                      'supported sizes are 64, 128')

    def set_pack_format(self):
        if self.size == DSize(64):
            self._pack_format = 'ff'
        elif self.size == DSize(128):
            self._pack_format = 'dd'
        else:
            raise NotImplementedError(f'Unsupported format: {self.size} - '
                                      'supported sizes are 64, 128')

    def pack(self, val):
        return self.bytes_to_bits(struct.pack(self.pack_format,
                                              val.real,
                                              val.imag))

    def unpack(self, buffer):
        vals = struct.unpack(self.pack_format, self.bits_to_bytes(buffer))
        return vals[0] + vals[1] * 1j


class DTuple(CompoundDeltaType):
    """Fixed collection of items of different types."""

    def __init__(self, elems: Iterable[Union[Type, BaseDeltaType]]):
        super().__init__()
        elem_list = []

        for elem in elems:
            elem_list.append(as_delta_type(elem))

        self.elems = tuple(elem_list)
        if len(self) == 0:
            raise ValueError("You cannot have empty DTuple")

        self.size = sum((e.size for e in self.elems), DSize(0))

    def __str__(self):
        return f"({', '.join(map(str, self.elems))})"

    def __len__(self):
        return len(self.elems)

    def as_python_type(self):
        return Tuple[tuple([t.as_python_type() for t in self.elems])]

    def as_numpy_object(self, val):
        m = np.zeros(1, dtype=[(f"f{i}", t.as_numpy_type())
                               for i, t in enumerate(self.elems)])
        for i, (v, t) in enumerate(zip(val, self.elems)):
            if type(t) is DArray:
                m[0][i][0][:] = t.as_numpy_object(v)
            else:
                m[0][i] = t.as_numpy_object(v)
        return m

    def from_numpy_object(self, val):
        if type(val) is np.ndarray:
            # Get first row of array
            if len(val) != 1:
                logger.warning(
                    "Warning, NumPy array object for DTuple has length != 1")
            val = val[0]
        return tuple((t.from_numpy_object(v))
                     for t, v in zip(self.elems, val))

    def as_numpy_type(self):
        return np.dtype([(f'f{i}', t.as_numpy_type())
                         for i, t in enumerate(self.elems)])

    def __eq__(self, other):
        if type(other) is DTuple:
            return self.elems == other.elems
        return False

    def __hash__(self):
        return hash(self.elems)

    def flatten(self, vals):
        flatten_vals = []
        for v in vals:
            v_df_type = delta_type(v)
            if is_primitive(v_df_type):
                flatten_vals.append(v)
            else:
                flatten_vals += v_df_type.flatten(v)
        return flatten_vals

    def deflatten(self, flatten_vals):
        vals = []
        for e in self.elems:
            if is_primitive(e):
                # TODO pop from the end would be better
                vals.append(flatten_vals.pop(0))
            else:
                vals.append(e.deflatten(flatten_vals))

        # matches the type of unpack
        return tuple(vals)

    def pack(self, val):
        if not self.is_packable(val):
            raise DeltaTypeError(f'Data does not match the packing format\n'
                                 f'{val=}\n{self=}')

        if type(val) is np.ndarray:
            val = self.from_numpy_object(val)

        return b"".join([t.pack(v) for t, v in zip(self.elems, val)])

    def unpack(self, buffer):
        vals = []
        for t in self.elems:
            vals.append(t.unpack(buffer[:t.size.val]))
            buffer = buffer[t.size.val:]
        return tuple(vals)

    def is_packable(self, val):
        t = delta_type(val)
        if type(t) is DTuple and len(t.elems) == len(self.elems):
            if type(val) is np.ndarray:
                # Get first row of array
                if len(val) != 1:
                    logger.warning(
                        "NumPy array object for DTuple has length != 1")
                val = val[0]
            return all(elem.is_packable(v) for elem, v in zip(self.elems, val))
        return False

    def set_pack_format(self):
        self._pack_format = "".join((e.pack_format for e in self.elems))


class DRecord(CompoundDeltaType):
    """Records are ordered collections of key-value pairs, with the key
    being a string and the value being of a given type.

    Parameters
    ----------
    attrs_type
        Class defined via ``attrs`` library, see examples.

    Examples
    --------
    A data container class created with ``attrs`` used for keyworded
    data transfer:

    .. code-block:: python

        >>> import attr
        >>> import deltalanguage as dl

        >>> @attr.s(slots=True)
        ... class RecBI:
        ...
        ...     x: bool = attr.ib()
        ...     y: int = attr.ib()

        >>> dl.DRecord(RecBI)
        {x: DBool, y: DInt32}
    """

    def __init__(self, attrs_type):
        if not attr.has(attrs_type):
            raise DeltaTypeError('Please provide class with attrs attributes')
        super().__init__()
        self.attrs_type = attrs_type

        if hasattr(attrs_type, "__annotations__"):
            elem_list = [(name, as_delta_type(t))
                         for name, t in attrs_type.__annotations__.items()]
        else:
            # Used if attrs_type constructed via make_class
            elem_list = [(a.name, as_delta_type(a.type))
                         for a in attrs_type.__attrs_attrs__]

        if not elem_list:
            raise ValueError("You cannot have empty record types.")

        names = [n for (n, _) in elem_list]
        if len(names) != len(set(names)):
            raise ValueError("Matching keys.")

        self.elems = tuple(elem_list)
        self.size = sum((e.size for (_, e) in self.elems), DSize(0))

    def __str__(self):
        def record_inst_printer(field):
            return f"{field[0]}: {field[1]}"
        return f"{{{', '.join(map(record_inst_printer, self.elems))}}}"

    def __len__(self):
        return len(self.elems)

    def as_python_type(self):
        # if attrs_type has delta types, they won't be converted to python
        return self.attrs_type

    def as_numpy_object(self, val):
        m = np.zeros(1, dtype=[(name, t.as_numpy_type())
                               for name, t in self.elems])
        for name, t in self.elems:
            if type(t) is DArray:
                m[0][name][0][:] = t.as_numpy_object(getattr(val, name))
            else:
                m[0][name] = t.as_numpy_object(getattr(val, name))

        return np.rec.array(m)

    def from_numpy_object(self, val):
        if isinstance(val, np.ndarray):
            # Get first row of array
            if len(val) != 1:
                logger.warning(
                    "Warning, NumPy array object for DRecord has length != 1")
            val = val[0]
        return self.attrs_type(*[t.from_numpy_object(val[name])
                                 for name, t in self.elems])

    def as_numpy_type(self):
        return np.dtype((np.record, [(name, t.as_numpy_type())
                                     for name, t in self.elems]))

    def __eq__(self, other):
        if type(self) is type(other):
            return self.elems == other.elems
        return False

    def __hash__(self):
        return hash(self.elems)

    def flatten(self, vals):
        flatten_vals = []
        for v in attr.astuple(vals, retain_collection_types=True):
            v_df_type = delta_type(v)
            if is_primitive(v_df_type):
                flatten_vals.append(v)
            else:
                flatten_vals += v_df_type.flatten(v)
        return flatten_vals

    def deflatten(self, flatten_vals):
        vals = []
        for (_, e) in self.elems:
            if is_primitive(e):
                # TODO pop from the end would be better
                vals.append(flatten_vals.pop(0))
            else:
                vals.append(e.deflatten(flatten_vals))

        # matches the type of unpack
        return self.attrs_type(*vals)

    def pack(self, val):
        if not self.is_packable(val):
            raise DeltaTypeError(f'Data does not match the packing format\n'
                                 f'{val=}\n{self=}')

        if isinstance(val, np.ndarray):
            val = self.from_numpy_object(val)

        return b"".join(
            [t[1].pack(v)
             for t, v in
             zip(self.elems, attr.astuple(val, retain_collection_types=True))])

    def unpack(self, buffer):
        vals = []
        for _, t in self.elems:
            vals.append(t.unpack(buffer[:t.size.val]))
            buffer = buffer[t.size.val:]
        return self.attrs_type(*vals)

    def is_packable(self, val):
        if isinstance(val, np.ndarray):
            # Get first row of NumPy array and load into attrs_type object
            if len(val) != 1:
                logger.warning(
                    "Warning, NumPy array object for DRecord has length != 1")
            val = self.attrs_type(*val[0])
        if attr.has(val):
            if len(attr.astuple(val)) != len(self.elems):
                return False

            def packable_attrs(name, t):
                return hasattr(val, name) and t.is_packable(getattr(val, name))
            return all(packable_attrs(name, t) for name, t in self.elems)
        else:
            return False

    def set_pack_format(self):
        self._pack_format = "".join(
            (e.pack_format for (_, e) in self.elems))


class DUnion(CompoundDeltaType):
    """Container for several Deltaflow types.

    Parameters
    ----------
    elems : Iterable[Union[Type, BaseDeltaType]]
        Data types.

    Examples
    --------
    .. code-block:: python

        >>> import deltalanguage as dl

        >>> dl.DUnion([int, bool])
        <DBool | DInt32>

    Note that the order or elements does not matter:

    .. code-block:: python

        >>> dl.DUnion([int, bool]) == dl.DUnion([bool, int])
        True
    """

    def __init__(self, elems: Iterable[Union[Type, BaseDeltaType]]):
        super().__init__()
        self.elems = []

        # flatten DUnions and save only unique sorted elements
        for e in elems:
            if type(e) is DUnion:
                for ee in e.elems:
                    self._add_elem(as_delta_type(ee))
            else:
                self._add_elem(as_delta_type(e))

        if len(self.elems) == 0:
            raise DeltaTypeError("You cannot have empty DUnion")

        # 8 bits to encode meta data
        self.meta = DUInt(DSize(8))
        self.size = self.meta.size + max(e.size for e in self.elems)

    def __str__(self):
        elems_str = ' | '.join(map(str, self.elems))
        return '<' + elems_str + '>'

    def __len__(self):
        return len(self.elems)

    def as_python_type(self):
        constituent_py_types = [t.as_python_type() for t in self.elems]
        return Union[tuple(constituent_py_types)]

    def as_numpy_object(self, val):
        array = np.zeros(1, dtype=self.as_numpy_type())
        for i, t in enumerate(self.elems):
            if t.is_packable(val):
                if type(t) is DArray:
                    array[0][i][0][:] = t.as_numpy_object(val)
                else:
                    array[0][i] = t.as_numpy_object(val)
        return array

    def from_numpy_object(self, val):
        """
        .. warning::
            While ``DUnion`` objects can be converted to NumPy objects, they
            cannot be converted back. This is because NumPy unions do not
            specify which type is being used.
        """
        raise DeltaTypeError(
            "NumPy unions cannot be converted to Python types.")

    def as_numpy_type(self):
        # Unions are implemented as a NumPy array
        # where all elements have the same offset.
        return np.dtype({'names': [str(elem) for elem in self.elems],
                         'formats': [t.as_numpy_type() for t in self.elems],
                         'offsets': [0]*len(self.elems)})

    def _add_elem(self, elem: BaseDeltaType):
        if not elem in self.elems:
            self.elems.append(elem)
            self.elems.sort(key=str)

    def __eq__(self, other):
        if type(self) is type(other):
            return self.elems == other.elems
        return False

    def __hash__(self):
        return hash(tuple(self.elems))

    def flatten(self, vals):
        # no need, instead the same method from the identified type is used
        raise NotImplementedError

    def deflatten(self, flatten_vals):
        # no need, instead the same method from the identified type is used
        raise NotImplementedError

    def pack(self, val):
        if type(val) is np.ndarray:
            raise DeltaTypeError(
                "NumPy unions cannot be converted to Python types.")
        for i, elem in enumerate(self.elems):
            if elem.is_packable(val):
                val_df_type = self.elems[i]
                # append empty bytes so buffer is always the same
                pad_len = self.size - self.meta.size - val_df_type.size
                if pad_len:
                    padding = b'0' * (pad_len).val
                else:
                    padding = ''
                buffer = val_df_type.pack(val) \
                    + padding \
                    + self.meta.pack(i)

                if len(buffer) != self.size.val:
                    raise ValueError(
                        f'Buffer size mismatch: {val}, {self.size.val}')

                return buffer

        else:
            raise DeltaTypeError(f'DUnion does not support {val}')

    def unpack(self, buffer):
        val_df_type_idx = self.meta.unpack(buffer[-8:])
        val_df_type = self.elems[val_df_type_idx]

        # pick the appropriate number of bytes
        val = val_df_type.unpack(
            buffer[:self.elems[val_df_type_idx].size.val]
        )

        return val  # -> can be any supported type

    def is_packable(self, val):
        if type(val) is np.ndarray:
            return False
        return any(elem.is_packable(val) for elem in self.elems)

    def set_pack_format(self):
        # this method does not make sence here as the data format is encoded
        # in the meta byte
        pass


class DRaw(BaseDeltaType):
    """Wrapper for converting objects of one data type into raw bits.

    Note that :meth:`pack` and :meth:`unpack` assume that the value
    has already been converted into raw bits. Use :meth:`as_bits`
    to ensure this is the case when returning this type from a Python node,
    and :meth:`from_bits` when receiving this type as input to a Python node
    to guarantee you get the value rather than its raw binary representation.

    Attributes
    ----------
    base_type : BaseDeltaType
        The data type we are converting to bits.
    """

    def __init__(self, base_type: Union[Type, BaseDeltaType]):
        self.base_type = as_delta_type(base_type)

    def __eq__(self, other):
        if type(other) is DRaw:
            return self.base_type == other.base_type
        else:
            return False

    def __hash__(self):
        return hash(self.base_type)

    def __getattr__(self, item):
        return getattr(self.base_type, item)

    def __str__(self):
        return f"DRaw({str(self.base_type)})"

    def as_numpy_object(self, val):
        return self.base_type.as_numpy_object(val)

    def from_numpy_object(self, val):
        return self.base_type.from_numpy_object(val)

    def as_numpy_type(self):
        return self.base_type.as_numpy_type()

    def as_python_type(self):
        return self.base_type.as_python_type()

    def is_packable(self, val):
        return isinstance(val, int)

    def set_pack_format(self):
        self.base_type.set_pack_format()

    def pack(self, val):
        if not self.is_packable(val):
            raise DeltaTypeError(
                "DRaw.pack requires int input. Use DRaw.as_bits before pack.")
        return format(val, f"0{self.size.val}b").encode("ascii")

    def unpack(self, buffer):
        return int(buffer, 2)

    def as_bits(self, val: BaseDeltaType) -> int:
        """Converts a given value into an integer containing its binary
        representation. This is done using the base type's
        :py:meth:`pack<deltalanguage.data_types.BaseDeltaType.pack>` method.

        Parameters
        ----------
        val : BaseDeltaType
            The value to convert to bits

        Returns
        -------
        int
            The binary representation of val
        """
        return self.unpack(self.base_type.pack(val))

    def from_bits(self, bits: int) -> BaseDeltaType:
        """Converts a bit sequence back to the object those bits represent.
        This is done using the base type's
        :py:meth:`unpack<deltalanguage.data_types.BaseDeltaType.unpack>` method.

        Parameters
        ----------
        bits : int
            The binary representation of the value

        Returns
        -------
        BaseDeltaType
            The original description of the value.
        """
        return self.base_type.unpack(self.pack(bits))


class DOptional:
    """Wrapper class that is used to identify optional input for nodes.

    .. note::
        Only :py:class:`InPort<deltalanguage.wiring.InPort>` can be optional.
        :py:class:`OutPort<deltalanguage.wiring.OutPort>` does not have such
        property.

    Examples
    --------
    Create a node with 1 compulsory and 1 optional inputs:

    .. code-block:: python

        >>> import deltalanguage as dl

        >>> @dl.DeltaBlock()
        ... def foo(a: int, b: dl.DOptional(int)) -> int:
        ...     if b is None:
        ...         return a
        ...     else:
        ...         return a + b

    As usual this function can be used as a normal python function:

    .. code-block:: python

        >>> foo(5, None)
        5

        >>> foo(5, 6)
        11

    Now let's see how it works in the context of
    :py:class:`DeltaGraph<deltalanguage.wiring.DeltaGraph>`:

    .. code-block:: python

        >>> @dl.DeltaBlock()
        ... def bar() -> int:
        ...     # this node will imitate an absence of input
        ...     pass

        # helper node that saves the result
        >>> s = dl.lib.StateSaver(int, verbose=True)

        >>> with dl.DeltaGraph() as graph:
        ...     foo_out = foo(42, bar())
        ...     s.save_and_exit(foo_out) # doctest:+ELLIPSIS
        save_and_exit...

        >>> rt = dl.DeltaPySimulator(graph)
        >>> rt.run()
        saving 42

    So we can see that the node runs even the optional input is not provided.
    """

    def __init__(self, t: Union[Type, BaseDeltaType]):
        self.type = as_delta_type(t)

    def __eq__(self, other):
        return isinstance(other, DOptional) and self.type == other.type

    def __hash__(self):
        return self.type.__hash__() + 1

    @staticmethod
    def as_python_type():
        raise DeltaTypeError('DOptional is not a data type')


def is_primitive(t: BaseDeltaType) -> bool:
    return isinstance(t, PrimitiveDeltaType)


class ForkedReturn:
    """Wrapper class that is used when a node has multiple outputs.

    This is simply due to a restriction of python language, which permits only
    a single object returned by ``return``, note that several objects are
    bundled in a named tuple automatically.

    Thus this class is used in this case, as the Deltaflow language allows
    users to send a node's output to multiple destinations.

    Use :py:func:`make_forked_return` factory.

    Parameters
    ----------
    elems : Dict[str, Union[Type, BaseDeltaType]]
        Dictionary mapping port index to data type.
    """

    def __init__(self, elems: Dict[str, Union[Type, BaseDeltaType]]):
        elem_dict = {}

        for name, t in elems.items():
            df_t = as_delta_type(t)
            if isinstance(df_t, DOptional):
                raise DeltaTypeError('Outputs cannot be optional.')
            elem_dict[name] = df_t

        self.elem_dict = elem_dict

    @property
    def keys(self):
        return self.elem_dict.keys()

    @property
    def types(self):
        return self.elem_dict.values()

    def __repr__(self) -> str:
        return "ForkedReturn(" + \
            ", ".join(f"{k}:{v}" for k, v in self.elem_dict.items()) + \
            ")"

    def __eq__(self, other):
        if type(other) is ForkedReturn:
            return self.elem_dict == other.elem_dict
        return False


def make_forked_return(elems: Dict[str, Union[Type, BaseDeltaType]]):
    """Factory function used when a node sends different results to multiple
    destinations.

    Creates a type and class of return, used by the
    node to store multiple results in a single object.
    The individual ports then only send their specific part of the output
    to the destination nodes.
    This workaround is needed because python natively supports functions
    with only one output.

    Parameters
    ----------
    elems : Dict[str, Union[Type, BaseDeltaType]]
        Dictionary mapping port index to data type sent by that port.

    Returns
    -------
    ForkedReturn
        Type of return objects.
    NamedTuple
        Class of return objects.

    Examples
    --------
    The return type and class should be defined before a node using them
    is defined, thus this factory should be called first:

    .. code-block:: python

        >>> import deltalanguage as dl

        >>> TwoIntsT, TwoIntsC = dl.make_forked_return({'x': int, 'y': int})

        >>> @dl.DeltaBlock()
        ... def foo(a: int) -> TwoIntsT:
        ...     return TwoIntsC(a//2, a%2)

    It also allows us to reuse the same type and class in multiple nodes.
    Also users can send out only partial outputs, i.e. not fill up all
    fields in output:

    .. code-block:: python

        >>> @dl.DeltaBlock()
        ... def bar(a: int) -> TwoIntsT:
        ...     if a%2 == 0:
        ...         return TwoIntsC(a, None)
        ...     else:
        ...         return TwoIntsC(None, a)

    Wiring into a graph then has one distinction in comparison with nodes
    that have only one output, namely the outputs have to be called by their
    labels:

    .. code-block:: python

        >>> @dl.DeltaBlock(allow_const=False)
        ... def baz(a: int, b: int) -> dl.Void:
        ...     print(f"{a=}, {b=}")
        ...     raise dl.DeltaRuntimeExit

        >>> with dl.DeltaGraph() as graph:
        ...     foo_out = foo(20)
        ...     baz(a=foo_out.x, b=foo_out.y) # doctest:+ELLIPSIS
        baz...

        >>> rt = dl.DeltaPySimulator(graph)
        >>> rt.run()
        a=10, b=0
    """
    py_type_elems = []

    for name, t in elems.items():
        if isinstance(t, BaseDeltaType):
            t = t.as_python_type()
        py_type_elems.append((name, t))

    return ForkedReturn(elems), NamedTuple("ForkedReturnNT", py_type_elems)


def as_delta_type(t: Type) -> Union[BaseDeltaType, ForkedReturn, DOptional]:
    """Map a generic python type to the corresponding Deltaflow type.

    Parameters
    ----------
    t : Type
        Python data type.

    Returns
    -------
    Union[BaseDeltaType, ForkedReturn, DOptional]
        Instance of :py:class:`BaseDeltaType` subclass representing the
        Deltaflow type mapped to the provided Python type.


    .. note::
        Due to restrictions of python typing (python built-in types as
        well as types from ``typing`` library are quite restrictive)
        compound objects need to be analysed using
        :py:func:`delta_type`.


    Examples
    --------
    Simply call this function with any supported python type:

    .. code-block:: python

        >>> import deltalanguage as dl

        >>> dl.as_delta_type(int)
        DInt32

    Any complex or unsupported data type is recognized as an instance of
    singleton :py:class:`Top`:

    .. code-block:: python

        >>> dl.as_delta_type(dict)
        T
    """
    if isinstance(t, (BaseDeltaType, ForkedReturn, DOptional)):
        return t

    elif t is Void:
        return t

    elif t is None or t is type(None):
        raise DeltaTypeError('Please use Void instead.')

    elif t in (bool, np.bool_) or t is np.dtype(np.bool_):
        return DBool()

    elif t is complex:
        return DComplex()

    elif t is np.complex64 or t is np.dtype(np.complex64):
        return DComplex(DSize(64))

    elif t is np.complex128 or t is np.dtype(np.complex128):
        return DComplex(DSize(128))

    elif t is int:
        return DInt(DSize(32))

    elif t is np.int8 or t is np.dtype(np.int8):
        return DChar()

    elif t is np.int16 or t is np.dtype(np.int16):
        return DInt(DSize(16))

    elif t is np.int32 or t is np.dtype(np.int32):
        return DInt(DSize(32))

    elif t is np.int64 or t is np.dtype(np.int64):
        return DInt(DSize(64))

    elif t is np.uint8 or t is np.dtype(np.uint8):
        return DChar()

    elif t is np.uint16 or t is np.dtype(np.uint16):
        return DUInt(DSize(16))

    elif t is np.uint32 or t is np.dtype(np.uint32):
        return DUInt(DSize(32))

    elif t is np.uint64 or t is np.dtype(np.uint64):
        return DUInt(DSize(64))

    elif t is float:
        return DFloat(DSize(32))

    elif t is np.float32 or t is np.dtype(np.float32):
        return DFloat(DSize(32))

    elif t is np.float64 or t is np.dtype(np.float64):
        return DFloat(DSize(64))

    elif t in (str, np.string_):
        return DStr()

    elif type(t) is tuple and t[0] is np.string_:
        return DStr(DSize(t[1]))

    elif attr.has(t):
        return DRecord(t)

    elif isinstance(t, np.dtype):
        if len(t.fields) == 1:
            return DArray(t.fields['f0'][0].base,
                          DSize(t.fields['f0'][0].shape[0]))

        elif t.type is np.record:
            t = attr.make_class(t.name, {name: attr.ib(
                type=t.fields[name][0].base) for name in t.names})
            return DRecord(t)

        else:
            offsets = [f[1] for f in t.fields.values()]
            if offsets == [0 for _ in range(len(t.fields))]:
                return DUnion([t.fields[name][0].base for name in t.names])
            else:
                return DTuple([t.fields[name][0].base for name in t.names])

    elif t in (tuple, list):
        raise DeltaTypeError(f'Type {t} is too complex, use data_type(obj) '
                             f'instead of as_data_type(type(obj))')

    elif hasattr(t, '__origin__') and t.__origin__ in (tuple, list):
        # via Tuple and List
        raise DeltaTypeError(f'Type {t} is too complex, use data_type(obj) '
                             f'instead of as_data_type(type(obj))')

    return Top()


def delta_type(val: object) -> Union[BaseDeltaType, ForkedReturn, DOptional]:
    """Identifies the Deltaflow type of the object.

    In case of a compound type it will recursively investigate its components
    as well.


    Parameters
    ----------
    val: object
        Any python object.

    Returns
    -------
    Union[BaseDeltaType, ForkedReturn, DOptional]
        Deltaflow data type analogue.

    Examples
    --------
    Simply call this function with any supported python object:

    .. code-block:: python

        >>> import deltalanguage as dl

        >>> dl.delta_type(42)
        DInt32

    More complex items such as tuples are also recognised:

    .. code-block:: python

        >>> dl.delta_type((42, False, 0.73))
        (DInt32, DBool, DFloat32)

    Any objects whose type is not supported defaults to :py:class:`Top`:

    .. code-block:: python

        >>> dl.delta_type({'a': False})
        T
    """

    if val is None:
        raise DeltaTypeError('Please use Void.')

    elif type(val) in (bool,
                       np.bool_,
                       np.dtype(np.bool_),
                       complex,
                       np.complex64,
                       np.complex128,
                       np.dtype(np.complex64),
                       np.dtype(np.complex128),
                       int,
                       np.int8,
                       np.int16,
                       np.int32,
                       np.int64,
                       np.uint8,
                       np.uint16,
                       np.uint32,
                       np.uint64,
                       np.dtype(np.int8),
                       np.dtype(np.int16),
                       np.dtype(np.int32),
                       np.dtype(np.int64),
                       np.dtype(np.uint8),
                       np.dtype(np.uint16),
                       np.dtype(np.uint32),
                       np.dtype(np.uint64),
                       float,
                       np.float32,
                       np.float64,
                       np.dtype(np.float32),
                       np.dtype(np.float64)):
        # => DBool, DUInt, DInt, DFloat, DComplex
        return as_delta_type(type(val))

    elif type(val) in (str, np.string_):
        if len(val) == 1:
            return DChar()
        else:
            return DStr(DSize(len(val)))

    elif isinstance(val, Tuple):
        sub_types = tuple(delta_type(e) for e in val)
        return DTuple(sub_types)

    elif isinstance(val, List):
        sub_types = tuple(delta_type(e) for e in val)
        if sub_types[1:] == sub_types[:-1]:
            return DArray(sub_types[0], DSize(len(sub_types)))
        else:
            raise DeltaTypeError(f'Please provide an array, i.e. a list with '
                                 f'same type elements, {sub_types=}')

    elif isinstance(val, np.ndarray):
        # NumPy transformations
        if len(val.dtype) == 0:
            return DArray(val.dtype, DSize(val.shape[0]))

        elif val.dtype.type is np.record:
            # => DRecord
            t = attr.make_class(val.dtype.name, {name: attr.ib(
                type=val.dtype.fields[name][0]) for name in val.dtype.names})
            return DRecord(t)

        else:
            offsets = [f[1] for f in val.dtype.fields.values()]
            if offsets == [0 for _ in range(len(val.dtype))]:
                return DUnion([f[0] for f in val.dtype.fields.values()])
            else:
                return DTuple([f[0] for f in val.dtype.fields.values()])

    # => Top
    return as_delta_type(type(val))




from typing import List

import numpy as np
from qiskit.providers.aer.noise import (
    coherent_unitary_error,
    depolarizing_error,
    NoiseModel,
    phase_amplitude_damping_error,
    thermal_relaxation_error
)


class CustomNoiseModel(NoiseModel):
    """
    Class that inherits from Qiskit's NoiseModel class for easy contruction of
    noise models suited to the gateset currently defined in the HAL.
    """

    def __init__(self,
                 coherent_error_angles: List[float] = None,
                 depol_errors: List[float] = None,
                 damping_params: List[float] = None,
                 relaxation_params: List[float] = None,
                 ):
        """Creating the noise model object and specifying the used gateset All
        erros are assigend uniformly to the set of single-/two-qubit gates

        Parameters
        ----------
        coherent_error_angles : List[float]
            Angles of unitary over-rotaions for single, and two-qubit gates.
        depol_errors : List[float], optional
            Params of single- and two-qubit depolarization noise channels.
        damping_params : List[float]
            Amplitude damping, phase damping and excited state population
            of a general amplitude + phase damping channel.
        relaxation_params : List[float]
            t1 time, t2 time, timestep (unit of time) and excited state
            population of a thermal relaxation noise channel
            """

        # create "empty" noise model from the original Qiskit class
        # Need to specify the basis of all qiskit prebuilt gates that should be
        # noisy
        super().__init__(basis_gates=['h', 't', 'x', 'y', 's', 'sdg', 'z',
                                      'cx', 'cy', 'cz'])

        # list of qiskit labels of all single qubit gates specified in the HAL
        # For all single qubit gates defined used in the Qiskit HAL, the
        # corresponding label must be given here, so that errors are assigned
        # to that gate
        # Uppercase letters are custom gates defined in the Qiskit backend
        # while lowercase labels are default qiskit labels
        self._sq_gates = ['h', 'PiXY', 'PiYZ', 'PiZX', 'R', 'RX', 'RY', 'RZ',
                          's', 'sdg', 'sqrt_x', 'SX', 'SY', 't', 'X', 'Y', 'Z']
        # list of qiskit labels of all two qubit gates specified in the HAL
        self._tq_gates = ['cx', 'cy', 'cz']

        # defining matrix representations of Pauli gates
        self.gate2matrix_dict = {
            'X': np.array([[0, 1], [1, 0]]) + 0j,
            'Y': np.array([[0, -1j], [1j, 0]]),
            'Z': np.array([[1, 0], [0, -1]]) + 0j,
        }

        # adding coherent overrotations to each single and two-qubit gate
        # using Qiskit's unitary error. This will be applied every time the
        # original gate is executed in a circuit
        if coherent_error_angles is not None:

            assert isinstance(coherent_error_angles, list), \
                'Need list of Single-qubit and two-qubit gate errors'
            assert len(coherent_error_angles) == 2

            self.sq_angle, self.tq_angle = coherent_error_angles

            error_h = coherent_unitary_error(
                self._sq_rot(self.sq_angle, np.pi/4, 0))
            self.add_all_qubit_quantum_error(error_h, ['h'])

            error_x = coherent_unitary_error(
                self._sq_rot(self.sq_angle, np.pi/2, 0))
            self.add_all_qubit_quantum_error(error_x, ['X', 'RX', 'PiYZ',
                                                       'sqrt_x'])

            error_y = coherent_unitary_error(
                self._sq_rot(self.sq_angle, np.pi/2, np.pi/2))
            self.add_all_qubit_quantum_error(error_y, ['Y', 'RY', 'PiZX'])

            error_z = coherent_unitary_error(self._sq_rot(self.sq_angle, 0, 0))
            self.add_all_qubit_quantum_error(error_z,
                                             ['Z', 'RZ', 's', 'sdg', 't', 'PiXY'])

            error_sx = coherent_unitary_error(
                self._sq_rot(self.sq_angle, np.pi/2, np.pi/4))
            self.add_all_qubit_quantum_error(error_sx, ['SX'])

            error_sy = coherent_unitary_error(
                self._sq_rot(self.sq_angle, np.pi/2, -np.pi/4))
            self.add_all_qubit_quantum_error(error_sy, ['SY'])

            error_cx = coherent_unitary_error(
                self._contr_sq_rot(self.tq_angle, np.pi/2, 0))
            self.add_all_qubit_quantum_error(error_cx, ['cx'])

            error_cy = coherent_unitary_error(
                self._contr_sq_rot(self.tq_angle, np.pi/2, np.pi/2))
            self.add_all_qubit_quantum_error(error_cy, ['cy'])

            error_cz = coherent_unitary_error(
                self._contr_sq_rot(self.tq_angle, 0, 0))
            self.add_all_qubit_quantum_error(error_cz, ['cz'])

        # adding a depolarization channel
        if depol_errors is not None:

            assert isinstance(depol_errors, list), \
                'Need list of Single-qubit and two-qubit depolarization params'
            assert len(depol_errors) == 2

            self.sq_depol, self.tq_depol = depol_errors

            sq_depol_error = depolarizing_error(self.sq_depol, 1)
            self.add_all_qubit_quantum_error(
                sq_depol_error, self._sq_gates, warnings=False)

            tq_depol_error = depolarizing_error(self.tq_depol, 2)
            self.add_all_qubit_quantum_error(
                tq_depol_error, self._tq_gates, warnings=False)

        # adding an amplitude and/or phase damping channel
        if damping_params is not None:

            assert isinstance(damping_params, list), \
                'Need list of amplitude damping, phase damping and excited \
                     state population'
            assert len(damping_params) == 3

            self.amp_param, self.phase_param, self.amp_pop = damping_params

            damping_error = phase_amplitude_damping_error(self.amp_param,
                                                          self.phase_param,
                                                          self.amp_pop)
            self.add_all_qubit_quantum_error(
                damping_error, self._sq_gates, warnings=False)

        # adding a themral relaxation error
        if relaxation_params is not None:

            assert isinstance(relaxation_params, list), \
                'Need list of t1 time, t2 time, gate time and ex. state pop.'
            assert len(relaxation_params) == 4

            self.t1, self.t2, self.time, self.damp_pop = relaxation_params

            relaxation_error = thermal_relaxation_error(self.t1,
                                                        self.t2,
                                                        self.time,
                                                        self.damp_pop)
            self.add_all_qubit_quantum_error(
                relaxation_error, self._sq_gates, warnings=False)

        # add the 'unitary' label to the basis set for all the custom gates
        # in the HAL specified as a UnitaryGate object
        self.add_basis_gates(['unitary'])

    def _sq_rot(self, lamda, theta, phi):
        """build the matrix form of a single qubit rotation of angle lamda
           with direction set by theta, phi in spherical coord.

            Parameters
            ----------
            lamda : float
                rotation angle
            theta : float
                azimuthal angle of the rotation axis
            phi : float
                polar angle of the rotation axis

            Returns
            -------
            np.array[float]
                2x2 array of the single qubit rotation gate
        """

        unitary = np.cos(lamda/2)*np.eye(2) + 0j
        unitary += -1j*np.sin(lamda/2)*np.sin(theta)*np.cos(phi) \
            * (np.array([[0, 1], [1, 0]]) + 0j)
        unitary += -1j*np.sin(lamda/2)*np.sin(theta)*np.sin(phi) \
            * np.array([[0, -1j], [1j, 0]])
        unitary += -1j*np.sin(lamda/2)*np.cos(theta) *\
                                             (np.array([[1, 0], [0, -1]]) + 0j)

        return unitary

    def _contr_sq_rot(self, lamda, theta, phi):
        """matrix representation of controlled single qubit gates
        using qiskits indexing of the qubits

        Parameters
        ----------
        lamda : float
            rotation angle
        theta : float
            azimuthal angle of the rotation axis
        phi : float
            polar angle of the rotation axis

        Returns
        -------
        np.array[float]
            4x4 np.array of the matrix representation of the controlled rotation
            matrix
        """

        unitary = np.zeros((4, 4)) + 0j
        unitary[0::2, 0::2] = np.eye(2)
        unitary[1::2, 1::2] = self._sq_rot(lamda, theta, phi)

        return unitary



"""Testing for fork naming errors
"""

import unittest

from deltalanguage.data_types import make_forked_return
from deltalanguage.wiring import (DeltaGraph,
                                  placeholder_node_factory,
                                  DeltaBlock,
                                  DeltaMethodBlock,
                                  Interactive)

ForkT, Fork = make_forked_return(
    {
        'send': int,
        'y': int
    }
)


@DeltaBlock()
def forked_block() -> ForkT:
    return Fork(1, 2)


class ForkNameTestClass:

    @DeltaMethodBlock()
    def forked_method_block(self) -> ForkT:
        return Fork(1, 2)


@Interactive(in_params={}, out_type=ForkT, name="ForkedInteractive")
def interactive_forked(node):
    node.send(Fork(3, 4))


class NamingForkTest(unittest.TestCase):

    def test_forked_block_naming(self):
        """Test that checks a forked DeltaBlock raises an error if the fork
           names clash with node attributes
        """

        with self.assertRaises(NameError):
            with DeltaGraph():
                forked_block()

    def test_forked_placeholder_naming(self):
        """Test that checks a forked placeholder raises an error if the fork
           names clash with node attributes
        """
        with self.assertRaises(NameError):
            with DeltaGraph():
                p = placeholder_node_factory()
                p.specify_by_func(forked_block)

    def test_forked_method_block_naming(self):
        """Test that checks a forked DeltaMethodBLock raises an error if the 
           fork names clash with node attributes
        """
        with self.assertRaises(NameError):
            instance = ForkNameTestClass()
            with DeltaGraph():
                instance.forked_method_block()

    def test_forked_interactive_naming(self):
        """Test that checks a forked Interactive raises an error if the fork
           names clash with node attributes
        """
        with self.assertRaises(NameError):
            with DeltaGraph():
                interactive_forked.call()


if __name__ == "__main__":
    unittest.main()



from ._delta_graph import DeltaGraph
from ._node_classes.placeholder_node import PlaceholderNode
from ._node_classes.real_nodes import as_node


def placeholder_node_factory(*args, name=None, **kwargs) -> PlaceholderNode:
    """Node factory for for :py:class:`PlaceholderNode`.

    The main use case of such nodes is allowing us to create cycles
    in :py:class:`DeltaGraph` by allowing data dependencies to be resolved
    out of the order. The need for this step

    Parameters
    ----------
    args
        Nodes to create in-ports for if needed.
    name
        Name for the placeholder.
    kwargs
        Nodes to create in-ports for by kw if needed.

    Returns
    -------
    PlaceholderNode
        Constructed placeholder node.


    Examples
    --------
    In this example we see a simple cycle of 2 nodes.
    The placeholder is first used to provide an input to a new node, and then
    it is specified by :py:meth:`PlaceholderNode.specify_by_node`:

    .. code-block:: python

        >>> import deltalanguage as dl

        >>> @dl.DeltaBlock()
        ... def foo(a: int) -> int:
        ...     if a%2 == 0:
        ...         return a
        ...     else:
        ...         return -a

        >>> @dl.Interactive(in_params={"a": int}, out_type=int)
        ... def bar(node):
        ...     internal_memory = 0
        ...
        ...     for i in range(5):
        ...         node.send(i)
        ...         internal_memory += node.receive("a")
        ...
        ...     print("0 - 1 + 2 - 3 + 4 =", internal_memory)
        ...     raise dl.DeltaRuntimeExit

        >>> with dl.DeltaGraph() as graph:
        ...     p = dl.placeholder_node_factory()
        ...     b = bar.call(a=p)
        ...     p.specify_by_node(foo(b))

        >>> rt = dl.DeltaPySimulator(graph)
        >>> rt.run()
        0 - 1 + 2 - 3 + 4 = 2

    .. warning::
        It is very important to design graphs in such a way that the exit
        condition
        :py:class:`DeltaRuntimeExit<deltalanguage.runtime.DeltaRuntimeExit>`
        can always be reached.
        For graphs with cyclic dependency of nodes it means that at least
        one node should contain an internal state that would terminate the
        cycle and redirect the flow of data in the graph.
        Formally it means that the graph of nodes (which might be cyclic) has
        a representation as a graph of states (which must be acyclic).

        In the example above the state of the graph changes at each cycle,
        with a clear termination condition, i.e. when ``for`` loop terminates.

        In case if nodes have non-determinism, the designer of the graph
        might need to think about a backup exit plan, such as timeout or a
        maximum number of iterations.
        For instance, runtime simulators can be provided with the timeout
        value, which will shut down the simulation regardless of the obtained
        result, e.g. see
        :py:meth:`DeltaPySimulator.run<deltalanguage.runtime.DeltaPySimulator.run>`.

    Also users can define placeholders by a usual python functions
    of class methods via :py:meth:`PlaceholderNode.specify_by_func` and
    :py:meth:`PlaceholderNode.specify_by_method` respectively.

    .. TODO:: Add examples for both cases.
    """
    # Get an appropriate name
    if name is None:
        name = DeltaGraph.get_next_placeholder_name()

    # Get the active graph from the top of the graph stack
    graph = DeltaGraph.current_graph()

    # Check if arguments are nodes.
    # If not, put them in PyConstBodies in the current graph
    pos_input_nodes = [as_node(arg, graph) for arg in args]
    kw_input_nodes = {name: as_node(arg, graph)
                      for (name, arg) in kwargs.items()}

    # Use PlaceholderNode constructor and return result
    return PlaceholderNode(graph, name, *pos_input_nodes, **kw_input_nodes)



from .__about__ import (
    __license__,
    __copyright__,
    __url__,
    __contributors__,
    __version__,
    __doc__
)

# user-facing classes are exposed via __all__
from .data_types import *
from .logging import make_logger
from .runtime import *
from .wiring import *

# standard library
from . import lib



import warnings

import numpy as np

from qiskit import Aer, execute, QuantumCircuit
from qiskit.circuit.library import (HGate, SdgGate, SGate, TGate,
                                    XGate, YGate, ZGate)
from qiskit.extensions.unitary import UnitaryGate
from qiskit.providers.aer.noise import NoiseModel

from deltalanguage.data_types import DUInt, DSize

from . import IQuantumSimulator
from ..hal import Masks, Opcode, Shifts


# defining some custom gates that arise from twirling

class SxGate(UnitaryGate):
    def __init__(self):
        super().__init__(np.array([[0, 1], [1j, 0]]), label='SX')


class SyGate(UnitaryGate):
    def __init__(self):
        super().__init__(np.array([[0, 1j], [1, 0]]), label='SY')


class PiXY(UnitaryGate):
    """Pi-rotation with axis in x-y-plane gate class."""

    def __init__(self, angle):
        super().__init__(np.array([[0, -np.sin(angle) - 1j*np.cos(angle)],
                                   [np.sin(angle) - 1j*np.cos(angle), 0]]),
                         label='PiXY')


class PiYZ(UnitaryGate):
    """Pi-rotation with axis in y-z-plane gate class."""

    def __init__(self, angle):
        super().__init__(np.array([[np.cos(angle), -1j*np.sin(angle)],
                                   [1j*np.sin(angle), -1*np.cos(angle)]]),
                         label='PiYZ')


class PiZX(UnitaryGate):
    """Pi-rotation with axis in z-x-plane gate class."""

    def __init__(self, angle):
        super().__init__(np.array([[np.cos(angle), np.sin(angle)],
                                   [np.sin(angle), -1*np.cos(angle)]]),
                         label='PiZX')


class RX(UnitaryGate):
    """ rotation around x-axis with custom label for noise model"""

    def __init__(self, angle):
        super().__init__(np.array([[1j*np.cos(angle/2), np.sin(angle/2)],
                                   [np.sin(angle/2), 1j*np.cos(angle/2)]]),
                         label='RX')

# Some gates that already exist in qiskit need to redefined with a custom label
# so that they can be recognized by the noise model


class X(UnitaryGate):
    """ rotation around x-axis with custom label for noise model"""

    def __init__(self):
        super().__init__((np.array([[0, 1], [1, 0]]) + 0j), label='X')


class RY(UnitaryGate):
    """ rotation around y-axis with custom label for noise model"""

    def __init__(self, angle):
        super().__init__(np.array([[1j*np.cos(angle/2), -1j*np.sin(angle/2)],
                                   [1j*np.sin(angle/2), 1j*np.cos(angle/2)]]),
                         label='RY')


class Y(UnitaryGate):
    """ rotation around x-axis with custom label for noise model"""

    def __init__(self):
        super().__init__(np.array([[0, -1j], [1j, 0]]), label='Y')


class RZ(UnitaryGate):
    """ rotation around z-axis with custom label for noise model"""

    def __init__(self, angle):
        super().__init__(1j*np.array([[np.exp(-1j*angle/2), 0],
                                      [0, np.exp(1j*angle/2)]]), label='RZ')


class Z(UnitaryGate):
    """ rotation around x-axis with custom label for noise model"""

    def __init__(self):
        super().__init__((np.array([[1, 0], [0, -1]]) + 0j), label='Z')


class SqrtXGate(UnitaryGate):
    def __init__(self):
        super().__init__(
            np.array([[1j, 1], [1, 1j]])/np.sqrt(2), label='sqrt_x')


class QiskitQuantumSimulator(IQuantumSimulator):
    """Qiskit implementation of the IQuantumSimulator interface.

    Parameters
    ----------
    register_size : int
        Size of the qubit register.
    seed : int
        Seed for both the random transpiler and the measurement sampling.
    simulator_backend : AerBackend
        Specifies which qiskit simulator backend to use, possibilities are:
        QasmSimulator, StatevectorSimulator/UnitarySimulator
        (for debugging).
    noise_model : NoiseModel
        Qiskit NoiseModel object to apply when running a circuit.


    .. TODO::
        Add and example how to use it in a graph (without any runtime).
    """

    def __init__(self,
                 register_size: int = 16,
                 seed: int = None,
                 simulator_backend=Aer.get_backend('qasm_simulator'),
                 noise_model: NoiseModel = None
                 ):

        self._circuit = None
        self._simulator_backend = simulator_backend
        np.random.seed(seed)
        self._seed = seed

        # defaulted to 16 because the bitcode status return
        # has 16 bits assigned for measurement results.
        self._qubit_register_size = register_size

        # stores control qubits
        self._control_qubit_indices = []

        # assign projectq gate to each opcode
        self._parameterised_gate_dict = {
            Opcode['R'].value: RZ,
            Opcode['RX'].value: RX,
            Opcode['RY'].value: RY,
            Opcode['RZ'].value: RZ,
            Opcode['PIXY'].value: PiXY,
            Opcode['PIYZ'].value: PiYZ,
            Opcode['PIZX'].value: PiZX,
        }

        self._constant_gate_dict = {
            Opcode['H'].value: HGate,
            Opcode['S'].value: SGate,
            Opcode['SQRT_X'].value: SqrtXGate,
            Opcode['T'].value: TGate,
            Opcode['X'].value: X,
            Opcode['Y'].value: Y,
            Opcode['Z'].value: Z,
            Opcode['INVS'].value: SdgGate,
            # consecutive S and X gate needed for RC
            Opcode['SX'].value: SxGate,
            # consecutive S and Y gate needed for RC
            Opcode['SY'].value: SyGate,
        }

        if noise_model is not None:
            assert isinstance(noise_model, NoiseModel)
            self._noise_model = noise_model
        else:
            self._noise_model = NoiseModel()

    def apply_gate(self, gate, qubit_index: int, parameter: float = None):
        """Receives command information and implements the gate on the
        corresponding qubit.

        Parameters
        ----------
        gate : UnitaryGate
            Qiskit gate to be applied.
        qubit_index : int
            Index of qubit for gate to be applied to.
        parameter : float
            Angle of gate if parametrised.
        """
        if self._circuit is not None:

            if len(self._control_qubit_indices) == 0:  # single qubit gate

                if parameter is not None:
                    self._circuit.append(gate(parameter), [qubit_index])
                else:
                    self._circuit.append(gate(), [qubit_index])

            else:  # controlled gate

                if qubit_index in self._control_qubit_indices:
                    raise ValueError(
                        f"Target qubit {qubit_index} already set-up as " +
                        "control qubit!"
                    )

                control_number = len(self._control_qubit_indices)

                if gate == X:
                    gate = XGate
                elif gate == Y:
                    gate = YGate
                elif gate == Z:
                    gate = ZGate
                else:
                    warnings.warn("Noise not supported on cotrolled gates " +
                                  "besides CX, CY and CZ")

                gate_indices = self._control_qubit_indices + [qubit_index]

                if parameter is not None:
                    controlled_gate = gate(parameter).control(control_number)
                else:
                    controlled_gate = gate().control(control_number)

                self._circuit.append(controlled_gate, gate_indices)

                # reset control indices
                self._control_qubit_indices = []

    def accept_command(
        self,
        command: DUInt(DSize(32))
    ) -> DUInt(DSize(32)):

        op = command >> Shifts.OPCODE.value
        qubit_index = (command & Masks.QUBIT_INDEX.value)

        if qubit_index + 1 > self._qubit_register_size:
            raise ValueError(
                f"Qubit index ({qubit_index}) greater than " +
                f"register size ({self._qubit_register_size})!"
            )

        if op == Opcode["STATE_PREPARATION"].value:
            self._circuit = None

            self._circuit = QuantumCircuit(self._qubit_register_size)

        elif op == Opcode["STATE_MEASURE"].value:

            self._circuit.measure_all()

            job = execute(self._circuit, backend=self._simulator_backend,
                          optimization_level=0,
                          basis_gates=self._noise_model.basis_gates,
                          noise_model=self._noise_model,
                          seed_simulator=np.random.randint(self._seed), shots=1)

            result_dict = job.result().get_counts()

            outcome_string = list(result_dict.keys())[0]

            # Each measurement sent should have all valid flags.
            # Therefore valid mask added to the 16bit measurement bitcode.
            measurement_binary = Masks.VALIDS.value
            # convert binary string to int,
            # qubit index 0 is least significant bit in the binary expansion
            measurement_binary += int(outcome_string, 2)

            return measurement_binary

        elif op == Opcode["CONTROL"].value:

            # add qubit index to self.control_qubit_indices, store
            # in memory until a gate is called, which is run controlled
            # on these qubits.

            if qubit_index in self._control_qubit_indices:
                raise ValueError(
                    f"Qubit {qubit_index} already set-up as control qubit!"
                )

            if len(self._control_qubit_indices)+1 == self._qubit_register_size:

                raise ValueError(
                    "Too many control qubits for register size of " +
                    "f{self._qubit_register_size}!"
                )

            self._control_qubit_indices += [qubit_index]

        elif op in self._parameterised_gate_dict.keys():
            angle = (command & Masks.ARG.value) >> Shifts.ARG.value
            angle *= (2 * np.pi) / 1024
            gate = self._parameterised_gate_dict[op]

            self.apply_gate(gate, qubit_index, angle)

        elif op in self._constant_gate_dict.keys():
            gate = self._constant_gate_dict[op]
            self.apply_gate(gate, qubit_index)

        elif op == Opcode['ID'].value:
            pass

        else:
            raise TypeError(f"{op} is not a recognised opcode!")



import unittest

from deltalanguage.wiring import (DeltaGraph,
                                  PythonBody,
                                  PyFuncBody,
                                  RealNode)


class SelectBody(unittest.TestCase):
    """Tests associated with body selection system
    including how to exclude and prefer different bodies using tags.
    """

    def test_select_body_on_construction(self):
        """Test if single node bodies get their body auto-selected.
        Multi-body nodes do not undergo auto-selection.
        """
        graph = DeltaGraph()
        body_1 = PythonBody()
        test_node = RealNode(graph, [body_1])

        self.assertEqual(test_node.body, body_1)

        body_2 = PythonBody()
        test_node_2 = RealNode(graph, [body_1, body_2])

        with self.assertRaises(ValueError):
            test_node_2.body

    def test_selection_overide(self):
        """Test no override occurs without override == True
        """
        graph = DeltaGraph()
        body_1 = PythonBody(tags=["preferred_1"])
        body_2 = PythonBody(tags=["preferred_2"])
        test_node = RealNode(graph, [body_1, body_2])

        test_node.select_body(preferred=["preferred_1"])
        self.assertEqual(test_node.body, body_1)

        test_node.select_body(preferred=["preferred_2"], override=False)
        self.assertEqual(test_node.body, body_1)

        test_node.select_body(preferred=["preferred_2"])
        self.assertEqual(test_node.body, body_2)

    def test_selection_exclusion_preference(self):
        """Test that selection works correctly with respect to body order,
        exclusions and preferences
        """
        graph = DeltaGraph()
        body_1 = PythonBody(tags=["exclude_me", "prefer_me"])
        body_2 = PythonBody()
        body_3 = PythonBody(tags=["prefer_me"])
        test_node = RealNode(graph, [body_1, body_2, body_3])

        test_node.select_body(exclusions=["exclude_me"])
        self.assertEqual(test_node.body, body_2)

        test_node.select_body(exclusions=["exclude_me"],
                              preferred=["prefer_me"])
        self.assertEqual(test_node.body, body_3)

    def test_no_bodies_node(self):
        """Test if no-body nodes throw errors on body access
        """
        graph = DeltaGraph()
        test_node = RealNode(graph, [])

        with self.assertRaises(ValueError):
            test_node.select_body()
            test_node.body

        body_1 = PythonBody(tags=["exclude_me"])
        body_2 = PythonBody(tags=["exclude_me"])
        test_node = RealNode(graph, [body_1, body_2])

        with self.assertRaises(ValueError):
            test_node.select_body(exclusions=["exclude_me"])
            test_node.body

    def test_class_as_tag(self):
        """Test classes can be used as tags
        """
        graph = DeltaGraph()
        body_1 = PythonBody()
        body_2 = PyFuncBody(lambda: False)
        test_node = RealNode(graph, [body_1, body_2])

        test_node.select_body(preferred=[PyFuncBody])
        self.assertEqual(test_node.body, body_2)

        with self.assertRaises(ValueError):
            test_node.select_body(exclusions=[PythonBody])
            test_node.body


if __name__ == "__main__":
    unittest.main()



"""Class to represent a placeholder node that can be replaced with a real node.
"""

from __future__ import annotations
import logging
from typing import (TYPE_CHECKING,
                    Callable,
                    Dict,
                    Iterable,
                    List,
                    Optional,
                    Tuple)

from deltalanguage.logging import make_logger
from .._decorators import get_func_in_params_out_type
from .._node_templates import NodeTemplate

from .abstract_node import ForkedNode, ProxyNode
from .node_bodies import PyInteractiveBody
from .real_nodes import as_node, PythonNode

if TYPE_CHECKING:
    from .._node_templates import InteractiveBodyTemplate
    from .._delta_graph import DeltaGraph
    from .abstract_node import AbstractNode
    from .port_classes import InPort


log = make_logger(logging.WARNING, "PROXY NODES")


class PlaceholderNode(ProxyNode):
    """Node class to represent nodes that have yet to be specified.

    These nodes have to be specified in the context of :py:class:`DeltaGraph`,
    their main use case is creation of cyclic dependencies in the graph,
    i.e. loops.

    Use :py:class:`placeholder_node_factory` for construction in the
    context of :py:class:`DeltaGraph`.

    Parameters
    ----------
    graph : DeltaGraph
        Graph this node is a part of.
    name : str
        Name is used to access this placeholder when we want to replace
        it with a real node.
    args
        Nodes specified positionally that will be output to the node
        this placeholder is replaced by.
    kwargs
        Nodes specified via kw that will be output to the node
        this placeholder is replaced by.
    """

    def __init__(self,
                 graph: DeltaGraph,
                 name: str,
                 *args: Iterable[AbstractNode],
                 **kwargs):
        # No referee to give to super until node is specified
        super().__init__()
        self.graph = graph
        self.key = name
        # Register myself as a proxy under a given name
        self.graph.add_placeholder(self.key, self)

        # Store of nodes to be used to make in-ports at a later time.
        # Ignored if this placeholder is specified by
        # `specify_by_node` or `specify_by_process`.
        self.future_in_port_args: Iterable[AbstractNode] = args

        # Store of nodes to be used to make in-ports at a later time.
        # Ignored if this placeholder is specified by "specify_by_node".
        self.future_in_port_kwargs: Dict[str, AbstractNode] = kwargs

        # Store of port_name, port_destination, index pairs used to
        # create out-ports later on.
        # There are currently no out-ports as only in-ports are specified
        # at init time.
        self.future_out_ports: List[Tuple[InPort, Optional[str]]] = []

    def __getattr__(self, item):
        """Get forked node to represent that indexed access or the result of
        the normal attribute request.
        """
        fetched = getattr(self.referee, item, None)

        if fetched is None:
            return ForkedNode(self, item)

        return fetched

    def add_out_port(self, port_destination: InPort, index=None):
        """If this placeholder has been specified, we will add the out-port
        to the referee.

        If this placeholder is yet to be specified, we will store this data
        for later use.
        """
        if self.referee is None:
            self.future_out_ports.append((port_destination, index))
        else:
            self.referee.add_out_port(port_destination, index)

    def specify_by_func(self,
                        my_func: Callable,
                        allow_const: bool = True,
                        node_key: Optional[str] = None,
                        lvl: int = logging.ERROR):
        """Makes this placeholder no longer a placeholder, it acts like
        the node constructed using the given function.

        In-ports and out-ports on the placeholder get added to the new node.

        Parameters
        ----------
        my_func : Callable
            Function we will make into a node that this placeholder will
            represent.
        allow_const : bool
            Flag to specify if the node body can be constant.
        node_key : Optional[str]
            Keyword argument used for providing the node to the block, in case
            the user wants to debug sending & receiving messages in an
            interactive console.
            Note that this should only be used for debugging; for
            Deltaflow programs in production it should be sufficient to use the
            inputs and return values of a block for communications.
        lvl : int
            Logging level for the node.
            By default only error logs are displayed.
        """

        template = NodeTemplate.merge_deltablock(None,
                                                 my_func,
                                                 allow_const,
                                                 node_key,
                                                 name=self.key,
                                                 lvl=lvl)

        self.referee = template.call_with_graph(self.graph,
                                                *self.future_in_port_args,
                                                **self.future_in_port_kwargs)

        # Add the out ports needed for the specified node
        for destination, index in self.future_out_ports:
            self.add_out_port(destination, index)

        # remove myself from the list of active placeholders for my graph
        self.graph.placeholders.pop(self.key)

    def specify_by_process(self, process_template: InteractiveBodyTemplate,
                           **kwnodes):
        """Make this placeholder into a node with interactive body
        with the given process function, and the given type of inputs and 
        outputs.

        Parameters
        ----------
        process_template : InteractiveBodyTemplate
            Function of type (PythonNode -> None) packed with its
            input and output types.
        kwnodes
            Input nodes to this process specified by keyword.
        """
        if self.future_in_port_args:
            log.warning("Positional arguments dropped when specifying "
                        "placeholder as interactive node")

        self.referee = process_template.call(**kwnodes)

        # Add destination out-ports
        for destination, index in self.future_out_ports:
            self.add_out_port(destination, index)

        # remove myself from the list of active placeholders for my graph
        self.graph.placeholders.pop(self.key)

    def specify_by_method(self,
                          method_func: Callable,
                          instance: object,
                          node_key: Optional[str] = None,
                          lvl: int = logging.ERROR):
        """Makes this placeholder no longer a placeholder, it acts like the
        node constructed using the given function.

        In-ports and out-ports on the placeholder get added to the new node.

        Parameters
        ----------
        instance : object
            Instantiated object this nodes method func will run on.
        method_func : Callable
            Method function to run.
        node_key : Optional[str]
            Keyword argument used for providing the node to the block, in case
            the user wants to debug sending & receiving messages in an
            interactive console.
            Note that this should only be used for debugging; for
            Deltaflow programs in production it should be sufficient to use the
            inputs and return values of a block for communications.
        """
        template = NodeTemplate.merge_deltamethod(None, method_func, node_key,
                                                  name=self.key, lvl=lvl)

        self.referee = template.call_with_graph(self.graph,
                                                instance,
                                                *self.future_in_port_args,
                                                **self.future_in_port_kwargs)

        # Add the out ports needed for the specified node
        for destination, index in self.future_out_ports:
            self.add_out_port(destination, index)

        # remove myself from the list of active placeholders for my graph
        self.graph.placeholders.pop(self.key)

    def is_const(self):
        # Constantness does not propegate through placeholders
        # This makes somewhat sense as a loop of constants makes no sense
        return False

    def specify_by_node(self, node: AbstractNode):
        """Makes this placeholder no longer a placeholder, it acts like
        the given node where this placeholder has been used as input, i.e
        where this placeholder has an out-port: add this out-port to
        the given node.

        Parameters
        ----------
        node : AbstractNode
            Node this placeholder now represents.
        """
        self.referee = node

        if self.future_in_port_args or self.future_in_port_kwargs:
            # specify_by_func should be used instead if this occurs
            log.warning("Placeholder at %s is being specified by "
                        "node but it already has input nodes itself.",
                        self.key)

        # Add the out ports needed for the specified node
        for destination, index in self.future_out_ports:
            self.add_out_port(destination, index)

        # remove myself from the list of active placeholders for my graph
        self.graph.placeholders.pop(self.key)



"""Module for defining a DeltaGraph, this is a collection of connected
nodes and placeholders.
"""

from __future__ import annotations
from collections import OrderedDict
from copy import deepcopy
import inspect
from typing import TYPE_CHECKING, Dict, List, Set, Tuple, Union
import logging
import textwrap

import networkx as nx
import matplotlib.pyplot as plt
import numpy as np

from deltalanguage.data_types import make_forked_return
from deltalanguage.logging import make_logger
from ..data_types import (BaseDeltaType,
                          DeltaIOError,
                          DeltaTypeError,
                          DUnion,
                          Top)
from ._node_classes.abstract_node import AbstractNode
from ._node_classes.node_bodies import Latency, PyConstBody, PyInteractiveBody
from ._node_classes.real_nodes import OutPort
from ._node_templates import NodeTemplate

if TYPE_CHECKING:
    from ._node_classes.placeholder_node import PlaceholderNode
    from ._node_classes.real_nodes import RealNode


class DeltaGraph:
    """A collection of nodes and placeholders that make up a single DeltaGraph.

    Also responsible for statically controlling a stack of active graphs
    and naming placeholders.

    Parameters
    ----------
    name : str
        Graph's name.

    lvl : int
        The level at which logs are displayed.
        These are the same levels as in Python's ``logging`` package.
        By default only error logs are displayed.

    Examples
    --------
    Graph definition starts with defining individual nodes (or importing them
    from a library). Most importantly there must be a node that formally ends
    the evaluation of the graph by calling
    :py:class:`DeltaRuntimeExit<deltalanguage.runtime.DeltaRuntimeExit>`.
    Here is a node like this, it prints an incoming integer and
    calls for exit:

    .. code-block:: python

        >>> import deltalanguage as dl

        >>> @dl.DeltaBlock(allow_const=False)
        ... def print_and_exit(a: int) -> dl.Void:
        ...     print(a)
        ...     raise dl.DeltaRuntimeExit

    The graph definition uses python's ``with`` statement where one simply
    wires up outputs of nodes with inputs of the other nodes:

    .. code-block:: python

        >>> with dl.DeltaGraph() as graph:
        ...     print_and_exit(42) # doctest:+ELLIPSIS
        print_and_exit...

    Simply print the graph to see its content:

    .. code-block:: python

        print(graph)

    Note that this graph has 2 nodes, as a dummy node has been created for
    a constant.
    Also note that the name of each node has its index appended in the end:

    One can also use the interface with ``matplotlib`` to draw
    the same graph:

    .. code-block:: python

        graph.draw()

    In order to evaluate the graph it should be executed on a runtime simulator
    or a runtime.
    The Deltaflow language comes with a simple runtime simulator
    :py:class:`DeltaPySimulator<deltalanguage.runtime.DeltaPySimulator>`:

    .. code-block:: python

        >>> import deltalanguage as dl

        >>> rt = dl.DeltaPySimulator(graph)
        >>> rt.run()
        42

    Node inputs and outputs must be typed (in case of no output
    :py:class:`Void<deltalanguage.data_types.Void>` is used as above).
    To check the correctness of wiring manually one can call this method:

    .. code-block:: python

        >>> graph.check()
        True

    Attributes
    ----------
    global_stack : list
        Store of active graphs, used as a stack.
    placeholder_name_index : int
        Used to ensure that every placeholder node in the graph has a unique
        name.
    """

    global_stack = []
    placeholder_name_index: int = 0

    def __init__(self, name=None, lvl: int = logging.ERROR):
        self.name = name
        self.placeholders: Dict[str, PlaceholderNode] = {}
        self.nodes: List[RealNode] = []
        self.log = make_logger(lvl, "DeltaGraph")

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        if name is None:
            self._name = "main"
        else:
            self._name = name

    def __enter__(self):
        """Creation of the graph context, sets this graph as the active graph
        at the top of the graph stack.
        """
        DeltaGraph.global_stack.append(self)
        return self

    def __exit__(self, *args):
        """End of the graph context, pops from top of graph stack as this
        graph is no longer active.
        """
        if not self.select_bodies(override=False):
            self.log.info(f'Graph {self.name} contains a node without a body, '
                          'please define it before execution.')
        DeltaGraph.global_stack.pop()

    def __str__(self) -> str:
        ret = f"DeltaGraph[{self.name}] {{\n"
        ret += textwrap.indent(''.join(map(str, self.nodes)), prefix='    ')
        ret += "}\n"

        if self.placeholders:
            ret += f"There are placeholders!\n{self.placeholders.keys()}.\n"

        return ret

    def __repr__(self) -> str:
        return str(self)

    def flatten(self, other):
        """Graphs can be declared in different scopes but we want to then
        merge them and use them in one graph.
        To do this, connected graphs must be merged into one graph containing
        a list of all the nodes across all original graphs.
        This function is called when an out-port is added to a node where
        the destination graph is different from the origin graph.

        Parameters
        ----------
        other : DeltaGraph
            The other graph that must be merged into this one
        """
        # Merge node lists and set alien nodes graph to be this graph
        alien_nodes = other.nodes
        for alien_node in alien_nodes:
            self.nodes.append(alien_node)
            alien_node.graph = self

        # Merge placeholder dicts
        self.placeholders.update(other.placeholders)

    def _merge_out_ports(self, out_ports):
        """Merge a list of out ports into a single out-port, creating a
        splitter node as required. Needed when the same out-port appears
        multiple times on the same node.

        Parameters
        ----------
        out_ports : List[OutPort]

        Returns
        -------
        OutPort
            Single merged out port that outputs to splitter that then
            outputs to all original destinations.
        """
        if len(out_ports) == 1:
            return out_ports[0]

        elif len(out_ports) > 1:
            # name type and node should be the same across all given out-ports
            merged_name = out_ports[0].port_name
            merged_type = out_ports[0].port_type
            merged_node = out_ports[0].node
            for out_port in out_ports[1:]:
                if out_port.port_name != merged_name:
                    raise ValueError(f"Merged out-port names " +
                                     f"{str(merged_name)} and " +
                                     f"{str(out_port.port_name)} " +
                                     f"do not match.")
                if out_port.port_type != merged_type:
                    raise ValueError(f"Merged out-port types " +
                                     f"{str(merged_type)} and " +
                                     f"{str(out_port.port_type)} " +
                                     f"do not match.")
                if out_port.node != merged_node:
                    raise ValueError(f"Merged out-port node " +
                                     f"{str(merged_node)} and " +
                                     f"{str(out_port.node)}" +
                                     f"do not match.")
            all_destinations = [x.destination for x in out_ports]

            # create a body of the splitter node
            _reps = len(all_destinations)
            _SplitterT, _SplitterC = make_forked_return(
                {'out' + str(i): merged_type for i in range(_reps)}
            )

            def _splitter(to_split: merged_type) -> _SplitterT:
                return _SplitterC(*(deepcopy(to_split) for _ in range(_reps)))

            # Dynamically annotated function does not get signature set
            # properly, so to use normal interface we have to manually set it
            params = [inspect.Parameter('to_split',
                                        inspect.Parameter.POSITIONAL_OR_KEYWORD,
                                        annotation=merged_type)]
            _splitter.__signature__ = inspect.Signature(parameters=params,
                                                        return_annotation=_SplitterT)

            # If merged_node has forked output only get the relevant argument
            # TODO This duplicates PyConstBody.eval and should be refactored
            if out_ports[0].index is None:
                merged_node_output = merged_node
            else:
                merged_node_output = getattr(merged_node, out_ports[0].index)

            # Construct the node using NodeTemplate
            template = NodeTemplate.merge_deltablock(None,
                                                     _splitter,
                                                     merged_node.is_const(),
                                                     name='splitter',
                                                     latency=Latency(time=100),
                                                     lvl=merged_node.log.level)
            new_node = template.call_with_graph(self, merged_node_output)

            # send splitters outputs to the original destinations
            for i, dest in enumerate(all_destinations):
                new_node.add_out_port(dest, 'out' + str(i))

            # get in port of the splitter
            _in_port_name = list(new_node.in_ports.keys())[0]
            merged_destination = new_node.in_ports[_in_port_name]

            # return a new out port instead of original out_ports
            return OutPort(merged_name,
                           merged_type,
                           merged_destination,
                           merged_node)

        else:
            return None  # An empty list has been given so return None

    def do_automatic_splitting(self):
        """You cannot have multiple out-ports for the same piece of data in
        Deltaflow.
        Instead you have a single wire to a splitter node which outputs several
        separate wires.
        This code will automatically add splitter nodes to this graph
        where needed.
        """
        # For each node, group out-ports by their name in a dict
        for node in self.nodes:
            out_ports_dict: Dict[str, List[OutPort]] = {}
            for x in node.out_ports:
                if x.port_name in out_ports_dict:
                    out_ports_dict[x.port_name].append(x)
                else:
                    out_ports_dict[x.port_name] = [x]

            # Set this nodes out ports to a list of merged out ports
            node.out_ports = [self._merge_out_ports(x)
                              for x in out_ports_dict.values()]

    def check(self):
        """Check that the graph is correct.

        This is a 2-stage process that is outlined below:

            1. Check nodes and wires (data streams):

                - Nodes names are unique.
                - I/O-ports are unique
                - All non-optional inputs are provided. If an optional input is
                  not provided or output does not go anywhere then it is fine.

            2. Check typing of wires, see the impl and characterization tests.

        Returns
        -------
        bool
            ``True`` if correct.


        .. todo::
            Add extra checks of splitting/forking of wires in step 1.
        """
        # check node names
        nodes_names = [node.name for node in self.nodes]
        if len(set(nodes_names)) != len(nodes_names):
            raise DeltaIOError(f"nodes' names should be unique\n"
                               f"graph={self}")

        # check ports
        in_ports_all = [port
                        for node in self.nodes
                        for port in node.in_ports.values()]
        out_ports_all = [port
                         for node in self.nodes
                         for port in node.out_ports]
        out_ports_dest = [port.destination
                          for node in self.nodes
                          for port in node.out_ports]
        if len(set(in_ports_all)) != len(in_ports_all):
            raise DeltaIOError(f"in_ports should be unique\n"
                               f"graph={self}")
        if len(set(out_ports_all)) != len(out_ports_all):
            raise DeltaIOError(f"out_ports should be unique\n"
                               f"graph={self}")
        if len(set(out_ports_dest)) != len(out_ports_dest):
            raise DeltaIOError(f"out_ports destianations should be unique\n"
                               f"graph={self}")

        # check port names
        # note: out_ports_names can be not unique
        in_ports_names = [port.port_name for port in in_ports_all]
        out_ports_dest_names = [port.port_name for port in out_ports_dest]
        if len(set(in_ports_names)) != len(in_ports_names):
            raise DeltaIOError(f"in_ports' names should be unique\n"
                               f"graph={self}")
        if len(set(out_ports_dest_names)) != len(out_ports_dest_names):
            raise DeltaIOError(f"out_ports' destination names should be unique\n"
                               f"graph={self}")

        # nodes with interactive bodies cannot be portless,
        # as they block simulation, until scheduling is implemented
        for node in self.nodes:
            for body in node.bodies:
                if (isinstance(body, PyInteractiveBody)
                        and len(node.in_ports) == 0
                        and len(node.out_ports) == 0):
                    raise DeltaIOError(
                        f"Node {node} without any I/O has an interactive body "
                        f"{body}. At the moment this is not supported.\n"
                        "Please either remove this body or connect this node "
                        "to another one."
                    )

        # check missing inputs
        in_ports_unused = [port
                           for node in self.nodes
                           for port in node.in_ports.values()
                           if not port in out_ports_dest]
        for port in in_ports_unused:
            if not port.is_optional:
                raise DeltaIOError(f"Mandatory input is not provided\n"
                                   f"graph={self}\n{port=}")

        # check each channel typing
        for port in out_ports_all:
            try:
                self.check_wire(type_s=port.port_type,
                                type_r=port.destination.port_type)
            except DeltaTypeError as port_type_err:
                raise DeltaTypeError(
                    port_type_err.args[0] + f'\n{port=}') from port_type_err

        return True

    def select_bodies(self,
                      exclusions: List[object] = [],
                      preferred: List[object] = [],
                      override=True):
        """Shortcut to selecting bodies accross whole graph.
        Applies exclusion rules then selects the first valid body for each
        node, preferred bodies selected if available.

        Returns
        -------
        bool
            ``True`` if all nodes now have a ``body``.
        """
        all_selected = True
        for node in self.nodes:
            is_selected = node.select_body(exclusions, preferred, override)
            all_selected = all_selected and is_selected
        return all_selected

    @staticmethod
    def check_wire(type_s: BaseDeltaType, type_r: BaseDeltaType):
        """Check that ``type_s`` sent by :py:class:`OutPort` can by received
        as ``type_r`` by :py:class:`InPort`.

        In case when ``type_s`` is an instance of
        :py:class:`DUnion<deltalanguage.data_types.DUnion>`, it is
        enforces that each each type is received, i.e. strict typing.

        Returns
        -------
        bool
            ``True`` if okay.

        Examples
        --------
        Unit tests of this method characterise the behaviour.

        .. warning::

            Deltaflow data types implemented as class instances with defined
            size. In order to compare 2 types types use `__eq__` method.

        """
        if (not isinstance(type_s, BaseDeltaType)) \
                or (not isinstance(type_r, BaseDeltaType)):
            raise DeltaTypeError(f"Unsupported types\n{type_s=}\n{type_r=}")

        if type_r == type_s:
            return True

        if type_r == Top():
            return True

        if type(type_r) == DUnion and Top() in type_r.elems:
            return True

        raise DeltaTypeError(f"Broken strict typing\n"
                             f"{type_s=}\n{type_r=}")

    def find_node_by_name(self, name: str):
        """Get the node(s) with the given name from the graph.

        Parameters
        ----------
        name : str
            Name of node to return

        Returns
        -------
        Union[AbstractNode, List[AbstractNode]]
            Node or list of nodes matching this name.
        """
        nodes = []
        for node in self.nodes:
            if not node.name.split('_')[-1].isdigit():
                raise ValueError(f'Node {node.name} has a strange name')

            if '_'.join(node.name.split('_')[:-1]) == name:
                nodes.append(node)

        if len(nodes) == 0:
            return None
        elif len(nodes) == 1:
            return nodes[0]
        else:
            return nodes

    def add_node(self, node: RealNode):
        """Add a node to this graph.

        Parameters
        ----------
        node : RealNode
        """
        self.nodes.append(node)

    def add_placeholder(self, key: str, placeholder: PlaceholderNode):
        """Add a placeholder with unique key to this graph.

        Parameters
        ----------
        key : str
            Key to add placeholder at.
        placeholder : PlaceholderNode
            Placeholder node to add.
        """
        if key in self.placeholders.keys():
            raise KeyError(
                f"Placeholder with name {key} already exists in current graph."
            )
        else:
            self.placeholders[key] = placeholder

    def get_networkx_graph(self, port_nodes: bool = False):
        """Construct NetworkX representation of graph.

        Parameters
        ----------
        port_nodes : bool
            If ``True``, ports are presented as types of nodes in the graph.
            If ``False``, port details are added as extra labels to edges.

        Returns
        -------
        nx.MultiDiGraph
            Directional multigraph object used for graphical representation of
            :py:class:`DeltaGraph`.
        """
        graph = nx.MultiDiGraph()
        for node in self.nodes:
            graph.add_node(node.name, name=node.name, type="block")
            for out_port in node.out_ports:
                if port_nodes:
                    # If output is forked, add node for out_port
                    # Else simply use main node
                    if out_port.index:
                        src = f"out_{str(out_port.port_name)}"
                        graph.add_node(src,
                                       name=out_port.index,
                                       type="out_port")
                        graph.add_edge(node.name, src)
                    else:
                        src = node.name
                    dest = f"in_{str(out_port.dest_port_name)}"
                    graph.add_node(dest,
                                   name=out_port.dest_port_index,
                                   type="in_port")
                    graph.add_edge(dest, out_port.dest_node_name)
                    graph.add_edge(src, dest, type=str(out_port.port_type))
                else:
                    if out_port.index:
                        src = out_port.index
                    else:
                        src = ""
                    dest = out_port.dest_port_index
                    graph.add_edge(node.name,
                                   out_port.dest_node_name,
                                   src=src,
                                   type=str(out_port.port_type),
                                   dest=dest)

        return graph

    def draw(self, seed: int = None, iterations: int = 50, **kwargs):
        """Draw the graph using NetworkX's default drawing tool.
        Additional parameters are passed to draw_networkx.
        See the NetworkX documentation for further details on parameters.
        """
        if seed is None:
            seed = np.random

        graph = self.get_networkx_graph()
        pos = nx.spring_layout(graph, seed=seed, iterations=iterations)

        # Check to see if there are duplicate edges
        if any([key >= 1 for (_, _, key) in graph.edges]):
            self.log.warning("Where multiple outputs from one node go to" +
                             "the same node, only one edge is visible")

        # Make figure size scale with node number
        fig_len = max(4, round(graph.number_of_nodes()/2))
        plt.figure(figsize=(round(fig_len*1.5), fig_len))

        # Add labels as matplotlib text objects
        for name, loc in pos.items():
            plt.text(loc[0], loc[1], name,
                     horizontalalignment='center', verticalalignment='center')

        plt.axis('off')
        nx.draw_networkx(graph, pos=pos, with_labels=False,
                         connectionstyle='arc3, rad = 0.1', **kwargs)

    def write_gml(self, path, **kwargs):
        """Write graph visualisation as a GraphML file.

        Notes
        -----
        With this file, we can now import the graph into a program such as
        [Cytoscape](https://cytoscape.org/). After importing the graph,
        Cytoscape allows us to apply a layout algorithm to this graph,
        or let it use the one it finds most suitable. We can also control
        the style of the graph, and modify the appearance of nodes and edges,
        including different styles for ports and nodes. An example style
        for Deltaflow graphs is given in `deltalanguage/DeltaStyle.xml`.

        After calling this method, with a path pointing to an empty .gml
        file, install and open Cytoscape.
        Select your .gml file from the import dialog via:
        File > Import > Network from file
        Ensure the 'Network' panel is selected on the left toolbar, in which
        you can select which network/graph you are visualising.
        To import a style select your .xml file via:
        File > Import > Style from file
        Along the left toolbar, select the 'Style' panel.
        In this panel there is a drop-down menu that allows you to pick
        which style you are editing or using to visualise graphs.
        To spread nodes out, you can apply a layout function.
        Layout -> Edge-weighted Spring Embedded Layout -> (none)
        """
        nx.write_gml(self.get_networkx_graph(**kwargs), path)

    @classmethod
    def get_next_placeholder_name(cls) -> str:
        """Class method to get a new, globally unique, placeholder name.

        Returns
        -------
        str
        """
        ret_val = "placeholder_" + str(cls.placeholder_name_index)
        cls.placeholder_name_index += 1
        return ret_val

    @classmethod
    def stack(cls):
        """Returns the graph stack."""
        return cls.global_stack

    @classmethod
    def current_graph(cls):
        """Returns the active graph at the top of the graph stack."""
        return cls.global_stack[-1]

    @classmethod
    def clean_stack(cls):
        """Fresh start the stack.

        This is useful for tests where a fresh start is needed.
        """
        cls.global_stack = []
        cls.placeholder_name_index = 0
        AbstractNode.reset_index()
